

"""Module for Feature IO classes.

Copyright (c) 2014, Ying Jin <yjin@cshl.edu >


This code is free software; you can redistribute it and/or modify it
under the terms of the Artistic License (see the file COPYING included
with the distribution).

@author:  Ying Jin
@contact: yjin@cshl.edu
"""

# ------------------------------------
# python modules
# ------------------------------------
import re
import sys
import logging
import struct
from array import array
from random import sample as random_sample
from operator import itemgetter
from math import sqrt,ceil,floor
import gzip
from time import time


from TEToolkit.Constants import *

# ------------------------------------
# Misc functions
# ------------------------------------


# ------------------------------------
# Classes
# ------------------------------------

class PeakIO:
    """IO for peak information.

    """
    def __init__ (self):
        self.peaks = {}
    
    def add (self, chromosome, start, end, summit=None, 
             peak_height=None, number_tags=None, 
             pvalue=None, fold_enrichment=None, fdr=None):
        """items: (peak start,peak end, peak length, peak summit, peak
        height, number of tags in peak region, peak pvalue, peak
        fold_enrichment, fdr) <-- tuple type
        """
        if chromosome not in self.peaks:
            self.peaks[chromosome]=[]
        self.peaks[chromosome].append((start,end,end-start,summit,
                                       peak_height,number_tags,
                                       pvalue,fold_enrichment,fdr))

    def filter_pvalue (self, pvalue_cut ):
        peaks = self.peaks
        new_peaks = {}
        chrs = list(peaks.keys())
        chrs.sort()
        for chrom in chrs:
            new_peaks[chrom]=[p for p in peaks[chrom] if p[6] >= pvalue_cut]
        self.peaks = new_peaks

    def filter_fc (self, fc_low, fc_up=None ):
        """Filter peaks in a given fc range.

        If fc_low and fc_up is assigned, the peaks with fc in [fc_low,fc_up)
        
        """
        peaks = self.peaks
        new_peaks = {}
        chrs = list(peaks.keys())
        chrs.sort()
        if fc_up:
            for chrom in chrs:
                new_peaks[chrom]=[p for p in peaks[chrom] if p[7] >= fc_low and p[7]<fc_up]
        else:
            for chrom in chrs:
                new_peaks[chrom]=[p for p in peaks[chrom] if p[7] >= fc_low]
        self.peaks = new_peaks

    def total (self):
        peaks = self.peaks
        chrs = list(peaks.keys())
        chrs.sort()
        x = 0
        for chrom in chrs:
            x += len(peaks[chrom])
        return x
    
    def ave_fold_enrichment (self):
        peaks = self.peaks
        chrs = list(peaks.keys())
        chrs.sort()
        x = 0
        t = 0
        for chrom in chrs:
            x += len(peaks[chrom])
            for p in peaks[chrom]:
                t+=p[7]
        return t/x

    def max_fold_enrichment (self):
        """Return the maximum fc.
        
        """
        peaks = self.peaks
        chrs = list(peaks.keys())
        chrs.sort()
        x = 0
        for chrom in chrs:
            if peaks[chrom]:
                m = max([i[7] for i in peaks[chrom]])
                if m>x:
                    x=m
        return x
        
    
    def tobed (self):
        text = ""
        chrs = list(self.peaks.keys())
        chrs.sort()
        for chrom in chrs:
            for peak in self.peaks[chrom]:
                text+= "%s\t%d\t%d\n" % (chrom,peak[0],peak[1])
        return text

    def towig (self):
        text = ""
        chrs = list(self.peaks.keys())
        chrs.sort()
        for chrom in chrs:
            for peak in self.peaks[chrom]:
                text+= "%s\t%d\t%d\n" % (peak[0],peak[1])
        return text
        
    def init_from_dict (self, data):
        """Initialize the data from a dictionary. Improper assignment
        will damage the data structure.
        
        """
        self.peaks = {}
        chrs = list(data.keys())
        chrs.sort()
        for chrom in chrs:
            self.peaks[chrom]=[]
            a = self.peaks[chrom].append
            for i in data[chrom]:
                a(i)

    def merge_overlap ( self ):
        """peak_candidates[chrom] = [(peak_start,peak_end,peak_length,peak_summit,peak_height,number_cpr_tags)...]

        """
        peaks = self.peaks
        new_peaks = {}
        chrs = list(peaks.keys())
        chrs.sort()
        for chrom in chrs:
            new_peaks[chrom]=[]
            n_append = new_peaks[chrom].append
            prev_peak = None
            peaks_chr = peaks[chrom]
            for i in range(len(peaks_chr)):
                if not prev_peak:
                    prev_peak = peaks_chr[i]
                    continue
                else:
                    if peaks_chr[i][0] <= prev_peak[1]:
                        s_new_peak = prev_peak[0]
                        e_new_peak = peaks_chr[i][1]
                        l_new_peak = e_new_peak-s_new_peak
                        if peaks_chr[i][4] > prev_peak[4]:
                            summit_new_peak = peaks_chr[i][3]
                            h_new_peak = peaks_chr[i][4]
                        else:
                            summit_new_peak = prev_peak[3]
                            h_new_peak = prev_peak[4]
                        prev_peak = (s_new_peak,e_new_peak,l_new_peak,summit_new_peak,h_new_peak,peaks_chr[i][5]+prev_peak[5])
                    else:
                        n_append(prev_peak)
                        prev_peak = peaks_chr[i]
            if prev_peak:
                n_append(prev_peak)
        del peaks
        self.peaks = new_peaks
        return True

    def overlap_with_other_peaks (self, peaks2, cover=0):
        """Peaks2 is a PeakIO object or dictionary with can be
        initialzed as a PeakIO. check __init__ for PeakIO for detail.

        return how many peaks are intersected by peaks2 by percentage
        coverage on peaks2(if 50%, cover = 0.5).
        """
        peaks1 = self.peaks
        if isinstance(peaks2,PeakIO):
            peaks2 = peaks2.peaks
        total_num = 0
        chrs1 = list(peaks1.keys())
        chrs2 = list(peaks2.keys())
        for k in chrs1:
            if not chrs2.count(k):
                continue
            rl1_k = iter(peaks1[k])
            rl2_k = iter(peaks2[k])
            tmp_n = False
            try:
                r1 = next(rl1_k)
                r2 = next(rl2_k)
                while (True):
                    if r2[0] < r1[1] and r1[0] < r2[1]:
                        a = sorted([r1[0],r1[1],r2[0],r2[1]])
                        if float(a[2]-a[1]+1)/r2[2] > cover:
                            if not tmp_n:
                                total_num+=1
                                tmp_n = True
                    if r1[1] < r2[1]:
                        r1 = next(rl1_k)
                        tmp_n = False
                    else:
                        r2 = next(rl2_k)
            except StopIteration:
                continue
        return total_num
        
class PosCnt:
    def __init__(self,pos,cnt):
        self.pos = pos
        self.cnt = cnt
        
    def __repr__(self):
        return repr((self.pos, self.cnt))

        
class FWTrackII:
    """Fixed Width Locations Track class II along the whole genome
    (commonly with the same annotation type), which are stored in a
    dict.

    Locations are stored and organized by sequence names (chr names) in a
    dict. They can be sorted by calling self.sort() function.
    """
    def __init__ (self,filename = "", fw=0,anno=""):
        """fw is the fixed-width for all locations.
        
        """
        self.fw = fw
        self.__locations = {}           # locations
        self.__pluscounts = {}   # Ying read counts strand +
        self.__minuscounts = {}  # Ying read counts strand -
        self.__sorted = False
        self.total = 0                  # total tags
        self.__tagsize = 0
        self.__binTags = []
        self.annotation = anno   # need to be figured out
        self.__srcfile = filename

    def get_name(self):
        return self.__srcfile
    
    def clear_bins(self):
        del self.__binTags[:]
        
    def read_in_bins (self,species):

        size = 0.0
        idx_map = dict() #maps of chrom to index
        
        chrom_lengths = species_chrom_lengths[species]
        
       # for i in chrom_lengths.values() :
       #     size += i
        
        chrname = 'chr1'
        if species == 'dm3' :
            chrname = 'chr2L'
        
        size = chrom_lengths[chrname]    
        
        for i in range(int(ceil(size/BIN_SIZE))):
            self.__binTags.append(0)
            
        pos = 0
        idx_map[chrname] = pos

        #read the bed file  

                #    logging.warn("chrom %s does not exist in chrom_size file\n" %(c))
                #else :
        
        for poscnts in self.__pluscounts[chrname] :
            #start = poscnts.pos 
            start = floor(poscnts)
            cnt = poscnts - start
            if start == poscnts :
                start -= 1
                cnt = 1
            offset = idx_map[chrname]+floor(start/BIN_SIZE)
            #self.__binTags[int(offset)] += poscnts.cnt
            self.__binTags[int(offset)] += cnt
        
        for poscnts in self.__minuscounts[chrname] :
            #start = poscnts.pos
            start = floor(poscnts)
            cnt = poscnts - start
            if start == poscnts :
                start -= 1
                cnt = 1            
            offset = idx_map[chrname]+floor(start/BIN_SIZE)
            #self.__binTags[int(offset)] += poscnts.cnt
            self.__binTags[int(offset)] += cnt   
        
            
        
      
    def get_bins(self,idx):
        
        cnts = array('f')
        pos = 0
      
        # idx is a byte array, each bit represent an element of self.__binTags
        
        for i in range(len(idx)):
            a = idx[i]
            pos = i * 8
            for j in range(8):
                b = a >> j
                if b & 1 ==1 :
                    cnts.append(1.0 * self.__binTags[pos+j])
        
        return cnts
    
    # return indices of bins with values fall in the fraction
    def get_bins_idx(self,fraction):
        
        idx = array('B')
        for i in range(MAX_BIT): 
            idx.append(0x00)
       # qt  = robjects.r['quantile']
        #val = qt(robjects.FloatVector(self.__binTags),prob=fraction)
       
        for i in range(len(self.__binTags)):
            pos = int(floor(i/8))
            offset = i % 8
         #   if self.__binTags[i] <= val[0]  :
            if self.__binTags[i] > 0  :
                a = 1 << offset
                idx[pos] = idx[pos] | a
            
        return idx
    
    def get_all_bins_idx(self):
        
        idx = array('B')
        for i in range(MAX_BIT): 
            idx.append(0x00)
       
        for i in range(len(self.__binTags)):
            pos = int(floor(i/8))
            offset = i % 8
            if self.__binTags[i] > 0 :
                a = 1 << offset
                idx[pos] = idx[pos] | a
            
        return idx            

    def get_bin_rc_v2 (self,chrom,fragsize):
        
        #reads = array('f',(0,) * int(ceil(1.0 *chrsize/bin_size)))
        #sizeOfArray = len(reads)
        reads = []
        pos = []
        
        if chrom not in self.__fileList:
            logging.warn("No reads at chromosome %s. Skip!\n" %(chrom))
            return None
        else:
            fn = self.__fileList[chrom]
            try:
                fh = open(fn,'r')
            except IOError:
                logging.error("cannot open file of chromosome %s.\n" %(chrom))
                return None
            else:
                for line in fh:
                    line = line.strip()
                    items = line.split('\t')
                    if len(items) < 6 :
                        logging.warn("Format error at %s line: %s. Skip!\n" %(fn,line))
                    else:
                        start = int(items[1])
                        end = int(items[2]) 
                        strand = items[5]
                        w = 1
                        if len(items) > 6 : # weight for each tag
                            w = float(items[6])
                        # Shift read location by fragment size.
                        loc = 0
                        if strand == '+' :
                            loc = start + int(fragsize/2) 
                        else :
                            loc = end - int(fragsize/2) 
                        # Increment read count in bin.
                         
                        #binIdx = int(floor(1.0*loc/bin_size))   
                       
                        #if(binIdx >= sizeOfArray) :
                        #    logging.error("%s  %d array index out of range! species may not match." %(items[0],pos))
                        #    sys.exit(0)
                        #reads[pos] += w
                        
                        pos.append(loc)   
                       
                        reads.append(w)
                        
                fh.close()
                return (reads,pos)    


    def libsize(self):
        return self.total
    
    def add_loc (self, chromosome, fiveendpos, strand, cnt):
        """Add a location to the list according to the sequence name.
        
        chromosome -- mostly the chromosome name
        fiveendpos -- 5' end pos, left for plus strand, right for neg strand
        strand     -- 0: plus, 1: minus
        """
        #if not self.__locations.has_key(chromosome):
        if chromosome not in self.__pluscounts :
            #self.__locations[chromosome] = [array(BYTE4,[]),array(BYTE4,[])] # for (+strand, -strand)
            self.__minuscounts[chromosome] = [] #[0]*100000 #[array(BYTE4,[])]
            self.__pluscounts[chromosome] = [] #[0]*100000 #[array(BYTE4,[])]
        #self.__locations[chromosome][strand].append(fiveendpos)
        #poscnt = PosCnt(fiveendpos,cnt)        
        if strand == 0:  # Ying
        
            self.__pluscounts[chromosome].append(fiveendpos + cnt)
        else:
            self.__minuscounts[chromosome].append(fiveendpos + cnt)
        
        #self.total+=1
       
        self.total +=cnt # Ying total counts  

    def get_locations_by_chr (self, chromosome):
        """Return a tuple of two lists of locations for certain chromosome.

        """
      
        if chromosome in self.__pluscounts:
            
            return (self.__pluscounts[chromosome],self.__minuscounts[chromosome])
        else:
            raise Exception("No such chromosome name (%s) in TrackI object!\n" % (chromosome))
        
    def get_locations_by_chr_v3 (self, chromosome):
        """Return a tuple of two lists of locations for certain chromosome.

        """

        if chromosome in self.__pluscounts:
            tagpos = []
            cnts = []
            pre_pos = -1
            cnt_sum = 0
            for i in range(len(self.__pluscounts[chromosome])) :
                poscnt = self.__pluscounts[chromosome][i]
                pos = int(floor(poscnt))
                cnt = poscnt - pos
                if pos == poscnt :
                    pos -= 1
                    cnt = 1
                if pre_pos == -1 :
                    pre_pos = pos
                    cnt_sum += cnt
                else :
                    if pre_pos == pos :
                        cnt_sum += cnt
                    else :
                        tagpos.append(pre_pos)
                        cnts.append(cnt_sum)
                        pre_pos = pos
                        cnt_sum = cnt
            if pre_pos != -1 :
                tagpos.append(pre_pos)
                cnts.append(cnt_sum)
                    
            return (tagpos, cnts)
            #return (self.__pluscounts[chromosome],self.__pluscounts[chromosome])
        else:
            raise Exception("No such chromosome name (%s) in TrackI object!\n" % (chromosome))

    def get_locations_by_chr_v2 (self, chromosome):
        """Return a tuple of two lists of locations for certain chromosome.

        """

        if chromosome in self.__pluscounts:
            tagpos = []
            cnts = []
            pre_pos = -1
           # cnt_sum = 0
            for i in range(len(self.__pluscounts[chromosome])) :
                poscnt = self.__pluscounts[chromosome][i]
                pos = int(floor(poscnt))
                cnt = poscnt - pos
                if pos == poscnt :
                    pos -= 1
                    cnt = 1
 
                tagpos.append(pos)
                cnts.append(cnt)
                    
            return (tagpos, cnts)
            #return (self.__pluscounts[chromosome],self.__pluscounts[chromosome])
        else:
            raise Exception("No such chromosome name (%s) in TrackI object!\n" % (chromosome))
            
            
    def clean(self,chr):
        self.__pluscounts[chr] = []
        
    def get_counts_by_chr (self, chromosome,strand=0): #Ying
        """Return a tuple of two lists of locations for certain chromosome.

        """
        if strand ==0:
            if chromosome in self.__pluscounts:
                return self.__pluscounts[chromosome]
            else:
                raise Exception("No such chromosome name (%s) in TrackI object!\n" % (chromosome))
        else:
            if chromosome in self.__minuscounts:
                return self.__minuscounts[chromosome]
            else:
                #raise Exception("No such chromosome name (%s) in TrackI object!\n" % (chromosome))
                return tmp;
       
    def get_chr_names (self):
        """Return all the chromosome names stored in this track object.
        """
        l = list(self.__pluscounts.keys())
        l.sort()
        return l

    def length (self):
        """Total sequenced length = total number of tags * width of tag
        """
        return self.total*self.fw
    
    def setTsize(self,tsize):
        self.__tagsize = tsize
        
    def tsize(self):
        return self.__tagsize

    def sort (self):
        """Naive sorting for locations.
        
        """
        for k in list(self.__pluscounts.keys()):
            
            (tmparrayplus,tmparrayminus) = self.get_locations_by_chr(k)
            self.__pluscounts[k] = sorted(tmparrayplus)
            self.__minuscounts[k] = sorted(tmparrayminus)
            #tmparray = self.get_counts_by_chr(k, 0) #Ying
            #self.__pluscounts[k] = sorted(tmparray,key=lambda poscnt: poscnt.pos)
 
            if len(tmparrayplus) < 1:
                logging.warning("NO records for chromosome %s, plus strand!" % (k))
            #self.__locations[k][1] = sorted(tmparrayminus)
            #tmparray = []
            #tmparray = self.get_counts_by_chr(k, 1) #Ying
            #self.__minuscounts[k] = sorted(tmparray,key=lambda poscnt: poscnt.pos)
            if len(tmparrayminus) < 1:            
                logging.warning("NO records for chromosome %s, minus strand!" % (k))
        self.__sorted = True

    def filter_dup (self,maxnum):
        """Filter the duplicated reads.

        Run it right after you add all data into this object.
        """
        if not self.__sorted:
            self.sort()
        self.total = 0
        for k in self.__locations: # for each chromosome
            # + strand
            plus = self.__locations[k][0]
            if len(plus) <1:
                new_plus = []
            else:
                new_plus = array(BYTE4,[plus[0]])
                pappend = new_plus.append
                n = 1                # the number of tags in the current location
                current_loc = plus[0]
                for p in plus[1:]:
                    if p == current_loc:
                        n += 1
                        if n <= maxnum:
                            pappend(p)
                    else:
                        current_loc = p
                        pappend(p)
                        n = 1
                self.total +=  len(new_plus)

            # - strand
            minus = self.__locations[k][1]
            if len(minus) <1:
                new_minus = []
            else:
                new_minus = array(BYTE4,[minus[0]])
                mappend = new_minus.append
                n = 1                # the number of tags in the current location
                current_loc = minus[0]
                for p in minus[1:]:
                    if p == current_loc:
                        n += 1
                        if n <= maxnum:
                            mappend(p)
                    else:
                        current_loc = p
                        mappend(p)
                        n = 1
                self.total +=  len(new_minus)
            self.__locations[k]=[new_plus,new_minus]

    def merge_plus_minus_locations_naive (self):
        """Merge plus and minus strand locations
        
        """
        #for chrom in self.__locations.keys():
        for chrom in self.__pluscounts:
            #(plus_tags,minus_tags) = self.__locations[chrom]
            #self.__locations[chrom][0].extend(self.__locations[chrom][1])
            #self.__locations[chrom][0] = sorted(self.__locations[chrom][0])
            #self.__locations[chrom][1] = []
            #Ying
            self.__pluscounts[chrom].extend(self.__minuscounts[chrom])
            self.__pluscounts[chrom] = sorted(self.__pluscounts[chrom]) #,key=lambda x: x.pos)
            self.__minuscounts[chrom] = []

            

    def merge_plus_minus_locations (self):
        """Merge plus and minus strand locations.

        Tao: Amazingly, this function for merging two sorted lists is
        slower than merge_plus_minus_locations_naive which only
        concatenate the two lists then sort it again! I am so discouraged!
        """
        if not self.__sorted:
            self.sort()
        for chrom in self.__locations:
            (plus_tags,minus_tags) = self.__locations[chrom]
            new_plus_tags = array(BYTE4,[])
            ip = 0
            im = 0
            lenp = len(plus_tags)
            lenm = len(minus_tags)
            while ip < lenp and im < lenm:
                if plus_tags[ip] < minus_tags[im]:
                    new_plus_tags.append(plus_tags[ip])
                    ip += 1
                else:
                    new_plus_tags.append(minus_tags[im])
                    im += 1
            if im < lenm:
                # add rest of minus tags
                new_plus_tags.extend(minus_tags[im:])
            if ip < lenp:
                # add rest of plus tags
                new_plus_tags.extend(plus_tags[ip:])
                    
            self.__locations[chrom] = [new_plus_tags,[]]
            self.total += len(new_plus_tags)


    def sample (self, percent):
        """Sample the tags for a given percentage.

        Warning: the current object is changed!
        """
        self.total = 0
        for key in self.__locations:
            num = int(len(self.__locations[key][0])*percent)
            self.__locations[key][0]=array(BYTE4,sorted(random_sample(self.__locations[key][0],num)))
            num = int(len(self.__locations[key][1])*percent)
            self.__locations[key][1]=array(BYTE4,sorted(random_sample(self.__locations[key][1],num)))
            self.total += len(self.__locations[key][0]) + len(self.__locations[key][1])
            
    def __str__ (self):
        return self.__to_wiggle()
        
    def __to_wiggle (self):
        """Use a lot of memory!
        
        """
        t = "track type=wiggle_0 name=\"tag list\" description=\"%s\"\n" % (self.annotation)
        for k in self.__locations:
            if self.__locations[k][0]:
                t += "variableStep chrom=%s span=%d strand=0\n" % (k,self.fw)
                for i in self.__locations[k][0]:
                    t += "%d\t1\n" % i
            if self.__locations[k][1]:
                t += "variableStep chrom=%s span=%d strand=1\n" % (k,self.fw)
                for i in self.__locations[k][1]:
                    t += "%d\t1\n" % i
        return t

class FWTrackIII:
    """Fixed Width Locations Track class III along the whole genome
    (commonly with the same annotation type), which are stored in a
    dict.

    Locations are stored and organized by sequence names (chr names) in a
    dict. They can be sorted by calling self.sort() function.
    """
    def __init__ (self,fw=0,anno=""):
        """fw is the fixed-width for all locations.
        
        """
        self.fw = fw
        self.chr =""
        self.__locations = [array(BYTE4,[]),array(BYTE4,[])]         # locations
        self.__pluscounts = []   # Ying read counts strand +
        self.__minuscounts = []  # Ying read counts strand -
        self.__sorted = False
        self.total = 0                  # total tags
        self.annotation = anno   # need to be figured out

    def add_loc (self, fiveendpos, strand, cnt):
        """Add a location to the list according to the sequence name.
        
       
        fiveendpos -- 5' end pos, left for plus strand, right for neg strand
        strand     -- 0: plus, 1: minus
        """
        
        self.__locations[strand].append(fiveendpos)
        poscnt = PosCnt(fiveendpos,cnt)        
        if strand == 0:  # Ying
            self.__pluscounts.append(poscnt)
        else:
            self.__minuscounts.append(poscnt)
        
        #self.total+=1
        self.total +=cnt # Ying total counts  

    def get_locations(self):
        """Return a tuple of two lists of locations for certain chromosome.

        """
        
        return self.__locations
        

    def get_counts (self, strand): #Ying
        """Return a tuple of two lists of locations for certain chromosome.

        """
        if strand ==0:
            return self.__pluscounts[chromosome]
        else:
            return self.__minuscounts[chromosome]
            
       
    def get_chr_names (self):
        """Return all the chromosome names stored in this track object.
        """
        return self.chr

    def length (self):
        """Total sequenced length = total number of tags * width of tag
        """
        return self.total*self.fw

    def sort (self):
        """Naive sorting for locations.
        
        """
        for k in self.__locations:
            (tmparrayplus,tmparrayminus) = self.get_locations_by_chr(k)
            self.__locations[k][0] = sorted(tmparrayplus)
            tmparray = self.get_counts_by_chr(k, 0) #Ying
            self.__pluscounts[k] = sorted(tmparray,key=lambda poscnt: poscnt.pos)
            if len(tmparrayplus) < 1:
                logging.warning("NO records for chromosome %s, plus strand!" % (k))
            self.__locations[k][1] = sorted(tmparrayminus)
            tmparray = []
            tmparray = self.get_counts_by_chr(k, 1) #Ying
            self.__minuscounts[k] = sorted(tmparray,key=lambda poscnt: poscnt.pos)
            if len(tmparrayminus) < 1:            
                logging.warning("NO records for chromosome %s, minus strand!" % (k))
        self.__sorted = True

    def merge_plus_minus_locations_naive (self):
        """Merge plus and minus strand locations
        
        """
        #(plus_tags,minus_tags) = self.__locations
        self.__locations[0].extend(self.__locations[1])
        self.__locations[0] = sorted(self.__locations[0])
        self.__locations[1] = []
            #Ying
        self.__pluscounts.extend(self.__minuscounts)
        self.__pluscounts.sorted(self.__pluscounts,key=lambda x: x.pos)
        self.__minuscounts = []         

    def merge_plus_minus_locations (self):
        """Merge plus and minus strand locations.

        Tao: Amazingly, this function for merging two sorted lists is
        slower than merge_plus_minus_locations_naive which only
        concatenate the two lists then sort it again! I am so discouraged!
        """
        if not self.__sorted:
            self.sort()
        (plus_tags,minus_tags) = self.__locations
        new_plus_tags = array(BYTE4,[])
        ip = 0
        m = 0
        lenp = len(plus_tags)
        lenm = len(minus_tags)
        while ip < lenp and im < lenm:
           if plus_tags[ip] < minus_tags[im]:
               new_plus_tags.append(plus_tags[ip])
               ip += 1
           else:
               new_plus_tags.append(minus_tags[im])
               im += 1
        if im < lenm:
                # add rest of minus tags
          new_plus_tags.extend(minus_tags[im:])
        if ip < lenp:
               # add rest of plus tags
           new_plus_tags.extend(plus_tags[ip:])
                   
        self.__locations = [new_plus_tags, []]
        self.total += len(new_plus_tags)


    def sample (self, percent):
        """Sample the tags for a given percentage.

        Warning: the current object is changed!
        """
        self.total = 0
        for key in self.__locations:
            num = int(len(self.__locations[key][0])*percent)
            self.__locations[key][0]=array(BYTE4,sorted(random_sample(self.__locations[key][0],num)))
            num = int(len(self.__locations[key][1])*percent)
            self.__locations[key][1]=array(BYTE4,sorted(random_sample(self.__locations[key][1],num)))
            self.total += len(self.__locations[key][0]) + len(self.__locations[key][1])
            
    def __str__ (self):
        return self.__to_wiggle()
        
    def __to_wiggle (self):
        """Use a lot of memory!
        
        """
        t = "track type=wiggle_0 name=\"tag list\" description=\"%s\"\n" % (self.annotation)
        for k in self.__locations:
            if self.__locations[k][0]:
                t += "variableStep chrom=%s span=%d strand=0\n" % (k,self.fw)
                for i in self.__locations[k][0]:
                    t += "%d\t1\n" % i
            if self.__locations[k][1]:
                t += "variableStep chrom=%s span=%d strand=1\n" % (k,self.fw)
                for i in self.__locations[k][1]:
                    t += "%d\t1\n" % i
        return t
        



'''
Created on Oct 13, 2011

two normalization methods are supported.
sequence depth
bin correlation

@author: Ying Jin
'''
#from rpy2.robjects.packages import importr
#from rpy2.robjects import FloatVector 
#import rpy2.robjects as robjects
#import rpy2.robjects.lib.ggplot2 as ggplot2
import subprocess


from array import array
#import numpy as np

import logging
import sys
from TEToolkit.Constants import *
from TEToolkit.ShortRead.ParseBEDFile import *


#stats = importr('stats')
#grdevices = importr('grDevices')


#def normalize(method,treatment,control,chrlen_tbl) :
def normalize(method,treatment,tinput,control,cinput,species,prj_name) :
    
    
    if method == 'sd' : #sequence depth
        return( seq_depth(treatment,tinput,control,cinput))
    elif method == 'bc' : #bin correlation
        #return( bin_corr(treatment,control,chrlen_tbl))
        return( bin_corr(treatment,tinput,control,cinput,species,prj_name))
    

def seq_depth(list1,list1input,list2,list2input):
    
    max_size = 0.0
    
    for i in range(len(list1)):
        tsmp = list1[i]
        if tsmp.libsize() > max_size :
            max_size = tsmp.libsize()

   
    for i in range(len(list2)):
       
        if list2[i].libsize() > max_size :
            max_size = list2[i].libsize()
            
    if len(list1input) > 0 and  max_size < list1input[0].libsize() :
        max_size = list1input[0].libsize()
        
    if len(list2input) > 0 and max_size < list2input[0].libsize() :
        max_size = list2input[0].libsize()
            
    sf1 = []
    sf2 = []

    for i in range(len(list1)):
        sf1.append(1.0 * max_size/list1[i].libsize())
        
    if len(list1input) > 0 :
        sf1.append(1.0*max_size/list1input[0].libsize())
    for i in range(len(list2)):
        sf2.append(1.0 * max_size/list2[i].libsize())
    if len(list2input) > 0 :
        sf2.append(1.0*max_size/list2input[0].libsize())
    
    return (sf1,sf2)

def __binCorr2r(filename):
    rfhd=open(filename,'w')
    
    rfhd.write("#!/bin/env Rscript\n") 

    rfhd.write("args = commandArgs(TRUE)\n")

    rfhd.write("infile = args[1]\n")
    rfhd.write("outfile = args[2]\n")

    rfhd.write("d<-read.delim(infile,header=T,stringsAsFactors=F) \n")

    rfhd.write("colnum <- length(d) \n")

    rfhd.write("min_sf_idx = 1 \n")
    rfhd.write("min_sf = 1.0 \n")

    rfhd.write("for (i in 2:colnum) {\n")
   
    rfhd.write("lm.r<-lm(d[,1]~d[,i]-1)\n")

    rfhd.write("cur_sf <- as.numeric(lm.r$coefficients[1])\n")

    #rfhd.write("if (min_sf > cur_sf ) { \n")
    rfhd.write("min_sf = cur_sf;\n")
    #rfhd.write("min_sf_idx = i }\n")
    
    rfhd.write("png(outfile,height=5,width=5,res=500,units='in')\n")
    rfhd.write("plot(d[,1]~d[,2],xlab=colnames(d)[1],ylab=colnames(d)[2],log='xy') \n")
    rfhd.write("dev.off() }\n")
    rfhd.write("cat(min_sf)\n")
    rfhd.close()
    



def __output_count_tbl(list1,list2, smp1,smp2, fname):
    
    
    try:
        f = open(fname, 'w')
    except IOError:
        error("Cannot create report file %s !\n" % (fname))
        sys.exit(1)
    else:
        cnt_tbl = dict()
        header = ""+smp1+"\t"+smp2
        f.write(header+"\n")
        for i in range(len(list1)) :
            f.write(str(list1[i]) + "\t" +str(list2[i])+"\n")    
            
        f.close()
    
    return

#def bin_corr(list1,list2,chrlen_tbl):
def bin_corr(list1,list1input,list2,list2input,species,prj_name):
    
    tisize = len(list1input)
    cisize = len(list2input)
    tsize = len(list1) 
    csize = len(list2) 
    fnames = []
    
    #fnames2 = []
    sf1 = []
    sf2 = []
    
    
    for i in range(tsize):
        fnames.append(list1[i].get_name())
        list1[i].read_in_bins(species) # read  tags of treatment samples into bins
    
    if tisize > 0 :
        fnames.append(list1input[0].get_name())
        list1input[0].read_in_bins(species)
    
       
    
    for i in range(csize):
        fnames.append(list2[i].get_name())
       # list2[i].read_in_bins(chrlen_tbl) # read tags of control samples into bins
        list2[i].read_in_bins(species)

    if cisize > 0 :
        fnames.append(list2input[0].get_name())
        list2input[0].read_in_bins(species)       
    
    # merge bins to make a matrix, rows are coordinates and cols are smps
    # randomly select the first sample as a temporary reference
    (tot_reads,reads,sel_idx) = join_bins(list1,list1input,list2,list2input,fnames,0)
    
    for i in range(csize):
        list2[i].clear_bins()
        
    if cisize > 0 :
        list2input[0].clear_bins() 
    if tisize > 0 :
        list1input[0].clear_bins() 
        
    for i in range(tsize):
        list1[i].clear_bins()    
        
    #call linear regression function
 #   refsmp = FloatVector(reads[0])
 #   robjects.globalenv['A'] = refsmp
   
    min_sf_idx = 0
    min_sf = 1.0
   
    __binCorr2r("bin_corr.r")
    
    for i in range(tsize+tisize):
        tmpfname = "."+fnames[0]+"-"+fnames[i]
        __output_count_tbl(reads[0],reads[i],fnames[0],fnames[i],tmpfname)
        outfname = prj_name+"_"+fnames[0]+"vs"+fnames[i]+".png"
        if i != 0 :
            #msf = subprocess.check_output(["Rscript","bin_corr.r",tmpfname,outfname])
            msf = subprocess.Popen(["Rscript","bin_corr.r",tmpfname,outfname],stdout=subprocess.PIPE)
            msf = float(msf.communicate()[0])
        else :
            msf = 1
     #   subprocess.call(["rm","-f",tmpfname])
        if min_sf > msf :
            min_sf = msf
            min_sf_idx = i
        sf1.append(msf)
    
    for i in range(tsize+tisize,csize + cisize + tsize+tisize):
        tmpfname = "."+fnames[0]+"-"+fnames[i]
        outfname = prj_name+"_"+fnames[0]+"vs"+fnames[i]+".png"
        __output_count_tbl(reads[0],reads[i],fnames[0],fnames[i],tmpfname)

        #msf = subprocess.check_output(["Rscript","bin_corr.r",tmpfname,outfname])
        msf = subprocess.Popen(["Rscript","bin_corr.r",tmpfname,outfname],stdout=subprocess.PIPE)
        msf = float(msf.communicate()[0])
      #  subprocess.call(["rm","-f",tmpfname])
        if min_sf > msf :
            min_sf = msf
            min_sf_idx = i
        sf2.append( msf)


    #plot scatter plot
    #sf = plot_sf(tot_reads,reads,sel_idx,fnames,min_sf_idx)
    
    #return (sf1[0:tsize],sf2[tsize:(tsize+csize)])
    return (sf1,sf2)

         
def join_bins(list1,list1input,list2,list2input,fnames,ref_idx):
    
    # only part of the background bins are used for computing normalization factors
    fraction = 1.0 #0.75
    idx_merged = array('B') #use a bit array to reduce memory usage
    reads = [] #np.array(dtype='double')
    tot_reads = []
    sel_idx = []
    #init a bit array
    for i in range(MAX_BIT): 
        idx_merged.append(0x00)
        
    # union of indices that have reads >0
    for i in range(len(list1)):
        idx = list1[i].get_all_bins_idx()       
        for j in range(len(idx)):
            idx_merged[j] = idx_merged[j] | idx[j]
            
    if len(list1input) > 0 :
        idx = list1input[0].get_all_bins_idx()
        for j in range(len(idx)):
            idx_merged[j] = idx_merged[j] | idx[j]
            
        
    for i in range(len(list2)):
        idx = list2[i].get_all_bins_idx()
        for j in range(len(idx)):
            idx_merged[j] = idx_merged[j] | idx[j]

    if len(list2input) > 0 :
        idx = list2input[0].get_all_bins_idx()
        for j in range(len(idx)):
            idx_merged[j] = idx_merged[j] | idx[j]    
            
    #get total reads
#    for i in range(len(list1)):
#        r = list1[i].get_bins(idx_merged)
#        tot_reads.append(r)

#    if len(list1input) > 0 :
#        r = list1input[i].get_bins(idx_merged)
#        tot_reads.append(r)      
        
#    for i in range(len(list2)):
#        r = list2[i].get_bins(idx_merged)
#        tot_reads.append(r)
        
#    if len(list2input) > 0 :
#        r = list2input[i].get_bins(idx_merged)
#        tot_reads.append(r) 
    
    idx_tot = []
    idx_tot.extend(idx_merged)    
    #intersect bins that are in the selected fraction        
#    for i in range(len(list1)):
#        idx = list1[i].get_bins_idx(fraction)       
#        for j in range(len(idx)):
#            idx_merged[j] = idx_merged[j] & idx[j]
            
        
#    for i in range(len(list2)):
#        idx = list2[i].get_bins_idx(fraction)
#        for j in range(len(idx)):
#            idx_merged[j] = idx_merged[j] & idx[j]
            
#    for i in range(len(idx_tot)):
#        a = idx_tot[i]
#        b = idx_merged[i]
#        if a | b != 0 :
#            if a & 1 == 1 and b & 1 == 1:
#                sel_idx.append(1)
#            if a & 1 == 1 and b & 1 == 0:
#                sel_idx.append(0)
#            for j in range(7):
#                a = a >> 1
#                b = b >> 1
#                if a & 1 == 1 and b & 1 == 1:
#                    sel_idx.append(1)
#                if a & 1 == 1 and b & 1 == 0:
#                    sel_idx.append(0) 
                    
   
    # retrieve reads in the bins (intersections of all samples) 
    for i in range(len(list1)):
        r = list1[i].get_bins(idx_merged)
        reads.append(r)
        
        
    if len(list1input) > 0 :
        r = list1input[0].get_bins(idx_merged)
        reads.append(r)
        
    for i in range(len(list2)):
        r = list2[i].get_bins(idx_merged)
        reads.append(r)
    
    if len(list2input) >0 :
        r = list2input[0].get_bins(idx_merged)
        reads.append(r)
    
    if len(reads) == 0 :
        logging.error("normalization failed !")
        sys.exit(0)
       
    return (tot_reads,reads,sel_idx)

'''
def plot_sf(t_reads,reads,sel_idx,smps,ref_idx):
    
    sf1 = []
  
    # plotting code here
    refsmp = FloatVector(reads[ref_idx])
    robjects.globalenv['A'] = refsmp
    
    for i in range(len(reads)):
        robjects.globalenv['B'] = FloatVector(reads[i]) 
        res = stats.lm('A ~ B - 1')
        slm = robjects.r.summary(res)
        sf1.append(res[0][0])
        r_sqr = slm[8][0]
        
        if i != ref_idx :
            fname = smps[ref_idx]+"_vs_"+ smps[i] + ".png"
            grdevices.png(file=fname, width=512, height=512)
            
            xlab_desc = 'Reads per 10kbp bin of sample ' + smps[i]
            ylab_desc = 'Reads per 10kbp bin of sample ' + smps[ref_idx]
            tt = "Slope: " + str(round(res[0][0],2)) + "\tR-squared: " + str(round(r_sqr,2))
            
            d = {'ref': robjects.FloatVector(t_reads[ref_idx]), 
            'smp': robjects.FloatVector(t_reads[i]),'background':robjects.IntVector(sel_idx)}
            dataf = robjects.DataFrame(d)
            
            gp = ggplot2.ggplot(dataf)
            pp = gp + \
                 ggplot2.aes_string(y = 'ref', x = 'smp', col='factor(background)') + \
                 ggplot2.geom_point(size=2,shape=16) + \
                 ggplot2.stat_abline(intercept=0,slope=res[0][0],col='blue') + \
                 ggplot2.opts(title=tt) + \
                 ggplot2.scale_colour_discrete(name="",breaks = [0,1], labels = ['all','selected']) + \
                 ggplot2.scale_x_continuous(xlab_desc) + \
                 ggplot2.scale_y_continuous(ylab_desc)                  

  #          pp.plot()
 #           grdevices.dev_off()

#    return sf1
'''





"""Module for Feature IO classes.

Copyright (c) 2014, Ying Jin <yjin@cshl.edu >


This code is free software; you can redistribute it and/or modify it
under the terms of the Artistic License (see the file COPYING included
with the distribution).

@author:  Ying Jin
@contact: yjin@cshl.edu
"""

# ------------------------------------
# python modules
# ------------------------------------

import math 
import sys
import array
import operator

from TEToolkit.Constants import *

def normalizeMeans(meansIn):
    total = sum(meansIn)
    meansOut = [0]*len(meansIn)
    sys.stderr.write("total means = " + repr(int(total)) +"\n")    
    if total > 0 :
        meansOut = [1.0*x/total for x in meansIn]
   #     for i in range(len(means)) :
   #         means[i] = 1.0*means[i]/total

    #sys.stderr.write("after normalization total means = "+str(sum(meansOut))+"\n")
    return meansOut
    
def EMUpdate(meansIn, te_features,uniq_counts,multi_reads,estimatedReadLength):
    # reassign multi-reads proportionally to the relative abundance of each TE  
    te_counts = {}
    meansOut = [0] *len(meansIn)

    multi_counts = computeAbundances(meansIn,multi_reads)    
    sys.stderr.write("total multi counts = "+ repr(int(sum(multi_counts)))+"\n")   
    for tid in range(len(meansIn)) :
        tlen = te_features.getLength(tid)
        if tlen <0 :
            sys.stderr.write("Error in optimization: the TE does not exist!\n")
            raise
        effectiveLength = tlen - estimatedReadLength + 1
        if effectiveLength > 0 : 
            meansOut[tid] = (uniq_counts[tid] + multi_counts[tid])/(1.0*effectiveLength)
        else :
         #   sys.stderr.write("effective length is less than read length\n")
            meansOut[tid] = 0.0
        
                  
    meansOut = normalizeMeans(meansOut)
    sys.stderr.write("after normalization total means = "+repr(sum(meansOut))+"\n")
   
    return meansOut

#def sum_(v) :
#    total = 0.0
    
#    for vid in v :
#        total += vid
    
#    return total
    
### ????
def logLikelihood_(means) :
    
 #   likelihoods = [0.0] *(transcriptsForKmer_.size())
    
    #Compute the log-likelihood
 #   for kid in transcriptsForKmer :
        #[&likelihoods, &means, this](const BlockedIndexRange& range) ->void {
 #       kmerLikelihood = 0.0
 #       for tid in transcriptsForKmer_[kid] :
 #           kmerLikelihood += transcripts_[tid].binMers[kid] * (means[tid] / transcripts_[tid].length)
  #          if kmerLikelihood > 1e-20 :
  #              likelihoods[kid] = log(kmerLikelihood)
                
            #else :
                #  " has probability too low\n"
    return 2 #sum_(likelihoods)
    
def expectedLogLikelihood_(means, counts, te_features) :
    sampProbs = [0.0]*len(means)
    
    for tid in range(len(counts)) : 
        relativeAbundance = means[tid]
   #     sampProbs[tid] = te_features[tid] * relativeAbundance
    
  #  normalize_(sampProbs)
    
    if len(means) > 0 :
        return logLikelihood( sampProbs) / len(means)
    else :
        return 0.0
  
def absdiff_(v0, v1) :
    diff = 0.0
    #for i in range(len(v0)) :
    #    diff += math.fabs(v0[i]-v1[i])
    diff = sum(map(lambda x,y: abs(x-y),v0,v1))
    return diff

def dotProd_(u, v) :
    dot = 0.0
    dot = sum(map(operator.mul,u,v))
    #for i in range(len(u)) :
    #    dot += u[i] * v[i]
    
    return dot
        

def EMestimate(te_features,multi_reads,uniq_counts,multi_counts,numItr,estimatedReadLength):
    
    #estimate average read length
   # estimatedReadLength = averageReadLength(multi_reads)
    
    #density per base
    sys.stderr.write("multi-reads = %s " % repr(int(len(multi_reads))))
    means0 = []
    for tid in range(len(uniq_counts)) :
        tlen = te_features.getLength(tid)
        if tlen < 0 :
            sys.stderr.write("Error in optimization: the TE does not exist!\n")
            raise
        effectiveLength = tlen  - estimatedReadLength + 1
        #if effectiveLength < 0 :
        #    effectiveLength = 1
        if effectiveLength > 0 :
            means0.append(1.0*(uniq_counts[tid] + multi_counts[tid])/effectiveLength)
        else :
            #sys.stderr.write("effective length is less that read length mean0.\n")
            means0.append(0.0)
    
    # relative abundance    
    means0 = normalizeMeans(means0)
    sys.stderr.write("after normalization total means0 = "+repr(sum(means0))+"\n")
    '''
    /**
         * Defaults for these values taken from the R implementation of
         * [SQUAREM](http://cran.r-project.org/web/packages/SQUAREM/index.html).
         */
    '''
    minStep0 = 1.0
    minStep = 1.0
    maxStep0 = 1.0
    maxStep = 1.0
    mStep = 4.0
    nonMonotonicity = 1.0
    negLogLikelihoodOld = float("inf")
    negLogLikelihoodNew = float("inf")
    # Right now, the # of iterations is fixed, but termination should
    # also be based on tolerance
    
    #while feval < numItr :
    cur_iter = 0
    #for iter in range(0,numItr) :
    t_size = len(uniq_counts)
    r = [0]*t_size
    r2 = [0] * t_size
    v = [0]*t_size
    meansPrime = [0.0] * t_size
    #means2 = [0.0] * t_size
    #means1 = [0.0] * t_size
    outerIteration = 1
    while cur_iter < numItr :
        cur_iter += 1
        sys.stderr.write("SQUAREM iteraton [" + str(cur_iter) + "]\n")
        sys.stderr.write("1/3\n")
        try :
            means1 = EMUpdate(means0, te_features,uniq_counts,multi_reads,estimatedReadLength)
        except :
            sys.stderr.write("Error in EMupdate\n")
            raise
        
        if negLogLikelihoodOld != float("inf") :
            negLogLikelihoodOld = -expectedLogLikelihood_(means0)
        
        sys.stderr.write("2/3\n")
        try:
            means2 = EMUpdate(means1, te_features,uniq_counts,multi_reads,estimatedReadLength)
        except :
            sys.stderr.write("Error in EMupdate\n")
            raise
        
#        delta = absdiff_(means1, means2)
 #       sys.stderr.write("delta = " + delta + "\n")
        
        for tid in range(len(means0)) :
            r[tid] = means1[tid] - means0[tid]
            r2[tid] = means2[tid] - means1[tid]
            v[tid] = (means2[tid] - means1[tid]) - r[tid]
        #    sys.stderr.write("mean1 = "+str(means1[tid])+"\t means2= "+str(means2[tid])+"\t"+str(v[tid])+"\n")
        #sys.exit(0)   
   
        rNorm = math.sqrt(dotProd_(r,r))
        r2Norm = math.sqrt(dotProd_(r2,r2)) 
        vNorm = math.sqrt(dotProd_(v,v))
        rr = dotProd_(r,v)
        rvNorm = math.sqrt(abs(rr))

        if vNorm == 0 :
            means0 = means1
            sys.stderr.write("at iteration " + str(cur_iter) + " vNorm == 0 \n")
            break
        alphaS = rNorm / rvNorm
        alphaS = max(minStep, min(maxStep, alphaS))
        
        if rNorm < OPT_TOL :
            sys.stderr.write("rNome = OPT_TOL \n")
            break
        if r2Norm < OPT_TOL :
            sys.stderr.write("r2Nome = OPT_TOL \n")
            means0 = means2
            break
        
        for tid in range(len(means0)) :
            meansPrime[tid] = max(0.0, means0[tid] + 2*alphaS*r[tid] + (alphaS*alphaS)*v[tid])
        
        # Stabilization step
        if abs(alphaS - 1.0) > 0.01 :
            sys.stderr.write("alpha = " + repr(alphaS) + ".\n ")
            sys.stderr.write("Performing a stabilization step.\n")
            try :
                meansPrime = EMUpdate(meansPrime, te_features,uniq_counts,multi_reads,estimatedReadLength)
            except :
                sys.stderr.write("Error in EMupdate\n")
                raise
             #/** Check for an error in meansPrime **/
             #/** If there is **/
            #if nonMonotonicity == float("inf") :
            #    sys.stderr.write("......in nonMonotonicity 1.......\n")
            #    negLogLikelihoodNew = - 1.0 * expectedLogLikelihood_(meansPrime)
            #else :
            #    sys.stderr.write("......in nonMonotonicity 2.......\n")
            #    negLogLikelihoodNew = negLogLikelihoodOld
            
            #if negLogLikelihoodNew > negLogLikelihoodOld + nonMonotonicity :
                #exchange values
            #    sys.stderr.write("......in nonMonotonicity 3.......\n")
            #    meansPrime, means2 = means2, meansPrime
            #    negLogLikelihoodNew = - 1.0 * expectedLogLikelihood_(meansPrime)
                
                if alphaS == maxStep :
                    maxStep = max(maxStep0, maxStep/mStep)
                    alphaS = 1.0
                
        sys.stderr.write("alpha = " + repr(alphaS) + ", ")
        
        if alphaS == maxStep :
            maxStep = mStep * maxStep
        
        if minStep < 0 and alphaS == minStep :
            minStep = mStep * minStep
        
        meansPrime, means0 = means0, meansPrime
        
        if not math.isnan(negLogLikelihoodNew) :
            negLogLikelihoodOld = negLogLikelihoodNew
    
    if cur_iter >= numItr :
        sys.stderr.write("not converge.....\n")
    else :
        sys.stderr.write("converge at iteration " + str(cur_iter)+"\n")    
    new_multi_counts = computeAbundances(means0,multi_reads)
    return new_multi_counts

def computeAbundances(meansIn,multi_reads):
        
    size = len(meansIn)
    multi_counts = [0] * size
    
    sys.stderr.write("num of multi reads = "+repr(int(len(multi_reads)))+"\n")
    for kid in range(len(multi_reads)) :
        TE_transcripts = multi_reads[kid]
        totalMass = 0.0
        for tid in TE_transcripts :
              totalMass += meansIn[tid]
        
        if totalMass > 0.0 :
                  norm = 1.0 / totalMass
        else :
                  norm = 0.0
                  
        for tid in TE_transcripts :
              multi_counts[tid] += meansIn[tid] * norm
    
    sys.stderr.write("total multi counts = "+ repr(int(sum(multi_counts)))+"\n")
    return multi_counts



"""Module Description

    Copyright (c) 2014, Ying Jin <yjin@cshl.edu >


    This code is free software; you can redistribute it and/or modify it
    under the terms of the Artistic License (see the file COPYING included
    with the distribution).

    @author:  Ying Jin
    @contact: yjin@cshl.edu
    """
import sys, time,re
import logging
import gzip
from math import ceil,floor
import collections

from TEToolkit.IntervalTree import *

#Taken from HTSeq
class GFF_Reader( ):

   """Parse a GFF file

   Pass the constructor either a file name or an iterator of lines of a
   GFF files. If a file name is specified, it may refer to a gzip compressed
   file.

   Yields tuple of (gene_id,chrom,strand,start position,end position,type)

   """

   def __init__( self, filename, id_attribute):
      self.line_no = None
      self.filename = filename
      self.id_attribute = id_attribute
      self._re_attr_main = re.compile( "\s*([^\s\=]+)[\s=]+(.*)" )

   def __iter__( self ):
      self.line_no = 0
      if self.filename.lower().endswith( ( ".gz" , ".gzip" ) ):
            lines = gzip.open( self.filename )
      else:
            lines = open( self.filename )

      for line in lines:
          self.line_no += 1
          if line == "\n" or line.startswith('#'):
              continue
          ( seqname, source, feature, start, end, score, strand, frame, attributeStr ) = line.split("\t")
          id = self.__parse_GFF_attr_string(attributeStr,self.id_attribute)

          yield (id, seqname, strand, int(start), int(end), feature)

      lines.close()
      self.line_no = None

   def __parse_GFF_attr_string(self,attributeStr,id_interested) :


       for pairs in attributeStr.split(';') :
           if pairs.count('"') not in [0,2] :
               raise ValueError("The attribute string seems to contain mismatched quotes.")
           nv = self._re_attr_main.match(pairs)
           if not nv :
               raise ValueError("Failure parsing GFF attribute line.")
           val = nv.group(2)
           name = nv.group(1)
           if name == id_interested :
               return val
       return None

   def get_line_number_string( self ):
      if self.line_no is None:
            return "file %s closed" % self.filename

      else:
         return "line %d of file %s" % ( self.line_no, self.filename )


class GeneFeatures:
    """index of Gene annotations.
        """
    def __init__ (self,GTFfilename,stranded,feature_type,id_attribute):

        self.featureIdxs_plus = {}
        self.featureIdxs_minus = {}
        self.featureIdxs_nostrand = {}
        self.features = []

        self.read_features(GTFfilename,stranded,feature_type,id_attribute)


    # Reading & processing annotation files
    def read_features(self,gff_filename, stranded, feature_type, id_attribute) :

        #dict of dicts since the builtin type doesn't support it for some reason
        temp_plus = collections.defaultdict(dict)
        temp_minus = collections.defaultdict(dict)
        temp_nostrand = collections.defaultdict(dict)

        # read count of features in GTF file
        gff = GFF_Reader(gff_filename,id_attribute)  # (id, seqname, strand, int(start), int(end), feature)
        i = 0
        counts = 0
        try:
            for f in gff:
                if f[0] is None :
                    continue
                if f[5] == feature_type:
                    counts += 1
                    if stranded != "no" and f[2] == "." :
                        sys.stderr.write("Feature %s does not have strand information." % (f[0]))
                    try:
                        if f[2] == "."  :
                            temp_nostrand[f[1]][f[0]].append((f[3],f[4]))
                    except:
                        temp_nostrand[f[1]][f[0]] = [(f[3],f[4])]

                    try:
                        if f[2] == "+"  :
                            temp_plus[f[1]][f[0]].append((f[3],f[4]))
                    except:
                        temp_plus[f[1]][f[0]] = [(f[3],f[4])]

                    try:
                        if f[2] == "-" :
                            temp_minus[f[1]][f[0]].append((f[3],f[4]))
                    except KeyError:
                        temp_minus[f[1]][f[0]] = [(f[3],f[4])]

                    #save gene id
                    if f[0] not in self.features :
                        self.features.append(f[0])

                    i += 1
                    if i % 100000 == 0 :
                        sys.stderr.write("%d GTF lines processed.\n" % i)
        except:
            sys.stderr.write("Error occured in %s.\n" % gff.get_line_number_string())
            raise

        if counts == 0 :
            sys.stderr.write("Warning: No features of type '%s' found in gene GTF file.\n" % feature_type)

        #build interval trees
        for each_chrom in temp_plus:
            inputlist = []
            for each_gene in temp_plus[each_chrom]:
                    for (start,end) in temp_plus[each_chrom][each_gene]:
                        inputlist.append(Interval(each_gene,start,end))
            self.featureIdxs_plus[each_chrom] = IntervalTree(inputlist)


        for each_chrom in temp_minus:
            inputlist = []
            for each_gene in temp_minus[each_chrom]:
                for (start,end) in temp_minus[each_chrom][each_gene]:
                    inputlist.append(Interval(each_gene,start,end))
            self.featureIdxs_minus[each_chrom] = IntervalTree(inputlist)

        for each_chrom in temp_nostrand:
            inputlist = []
            for each_gene in temp_nostrand[each_chrom]:
                for (start,end) in temp_nostrand[each_chrom][each_gene]:
                    inputlist.append(Interval(each_gene,start,end))
            self.featureIdxs_nostrand[each_chrom] = IntervalTree(inputlist)


    def getFeatures(self) :
        return self.features

    def Gene_annotation(self,itv_list):
        genes = []
        #fs = []
        for itv in itv_list :
            fs = []
            try:
                if itv[3] == "+" :
                    if itv[0] in self.featureIdxs_plus :
                        fs = self.featureIdxs_plus[itv[0]].find_gene(itv[1],itv[2])


                if itv[3] == "-" :
                        if itv[0] in self.featureIdxs_minus:
                            fs = self.featureIdxs_minus[itv[0]].find_gene(itv[1], itv[2])


                if itv[3] == "." :
                        if itv[0] in self.featureIdxs_minus:
                            fs = self.featureIdxs_minus[itv[0]].find_gene(itv[1], itv[2])

                        if itv[0] in self.featureIdxs_plus :
                            fs += self.featureIdxs_plus[itv[0]].find_gene(itv[1],itv[2])
                        if itv[0] in self.featureIdxs_nostrand :
                            fs += self.featureIdxs_nostrand[itv[0]].find_gene(itv[1],itv[2])

                if len(fs) > 0:
                        genes = genes + fs

            except:
                    raise


        return genes





"""Module Description

Copyright (c) 2014, Ying Jin <yjin@cshl.edu >


This code is free software; you can redistribute it and/or modify it
under the terms of the Artistic License (see the file COPYING included
with the distribution).

@author:  Ying Jin
@contact: yjin@cshl.edu
"""
import sys
import time
import logging
from math import ceil, floor
from TEToolkit.Constants import TEindex_BINSIZE

sys.setrecursionlimit(10000)


class Node:
    def __init__(self, start=-1, end=-1, name=-1, parent=None, left=None, right=None):
        self.__start = start
        self.__end = end
        self.__name = name  # idx in nameIDmap
        self.__namelist = {}
        self.left = left
        self.right = right
        self.balanceFactor = 0
        self.parent = parent
        self.isroot = False
        self.add(start, end, name)

    def add(self, start, end, name):
        if start in self.__namelist:
            self.__namelist[start].append((name, end))
        else:
            self.__namelist[start] = [(name, end)]

    def isRoot(self):
        return self.isroot

    def isLeftChild(self):
        return self.parent and self.parent.left == self

    def isRightChild(self):
        return  self.parent and self.parent.right == self

    def getStart(self):
        bin_startID = self.__start//TEindex_BINSIZE
        if self.__start == bin_startID * TEindex_BINSIZE:
            bin_startID -= 1
        return bin_startID

    def getEnd(self):
        bin_endID = self.__end//TEindex_BINSIZE
        return bin_endID

    def getName(self):
        return self.__name

    def overlaps(self, start, end):
        TEnamelist = []
        for s in sorted(self.__namelist.keys()):
            if s > end:
                break
            eles = self.__namelist[s]
            for name, e in eles:
                if start <= e and end >= s:
                    TEnamelist.append(name)

        return TEnamelist


class BinaryTree:
    def __init__(self):

        self.root = None
        self.size = 0

    def children_count(self):
        """
        Return the number of children

        @returns number of children: 0, 1, 2
        """
        cnt = 0
        if self.left:
            cnt += 1
        if self.right:
            cnt += 1
        return cnt

    def insert(self, start, end, name):
        if self.root:
            self.__insert(self.root, start, end, name)
        else:
            self.root = Node(start, end, name)
            self.root.isroot = True
        self.size = self.size + 1

    def __insert(self, node, start, end, name):
        """
        Insert new node with data

        @param data node data object to insert
        """
        root = node
        binstart = start//TEindex_BINSIZE
        if start == binstart * TEindex_BINSIZE:
            binstart -= 1

        if binstart == root.getStart():
            root.add(start, end, name)

        if binstart < root.getStart():
            if root.left is None:
                root.left = Node(start=start, end=end, name=name, parent=root)

                self.updateBalance(root.left)
            else:
                self.__insert(root.left, start, end, name)
        if binstart > root.getStart():
            if root.right is None:
                root.right = Node(start, end, name, parent=root)

                self.updateBalance(root.right)
            else:
                self.__insert(root.right, start, end, name)

    def updateBalance(self, node):
        if node.balanceFactor > 1 or node.balanceFactor < -1:
            self.rebalance(node)
            return

        if node.parent is not None:
            if node.isLeftChild():
                node.parent.balanceFactor += 1
            elif node.isRightChild():
                node.parent.balanceFactor -= 1

            if node.parent.balanceFactor != 0:
                self.updateBalance(node.parent)

    def rebalance(self, node):
        if node.balanceFactor < 0:
            if node.right.balanceFactor > 0:
                self.rotateRight(node.right)
                self.rotateLeft(node)
            else:
                self.rotateLeft(node)
        elif node.balanceFactor > 0:
            if node.left.balanceFactor < 0:
                self.rotateLeft(node.left)
                self.rotateRight(node)
            else :
                self.rotateRight(node)

    def rotateRight(self, oldRoot):
        newRoot = oldRoot.left
        oldRoot.left = newRoot.right
        if newRoot.right is not None:
            newRoot.right.parent = oldRoot
        newRoot.parent = oldRoot.parent

        if oldRoot.isRoot():
            self.root = newRoot
            newRoot.isroot = True
        else:
            if oldRoot.isLeftChild():
                oldRoot.parent.left = newRoot
            else:
                oldRoot.parent.right = newRoot

        newRoot.right = oldRoot
        oldRoot.parent = newRoot
        oldRoot.balanceFactor = oldRoot.balanceFactor - 1 - max(newRoot.balanceFactor, 0)
        newRoot.balanceFactor = newRoot.balanceFactor - 1 + min(oldRoot.balanceFactor, 0)

    def rotateLeft(self, oldRoot):
        newRoot = oldRoot.right
        oldRoot.right = newRoot.left
        if newRoot.left is not None:
            newRoot.left.parent = oldRoot
        newRoot.parent = oldRoot.parent

        if oldRoot.isRoot():
            self.root = newRoot
            newRoot.isroot = True
        else:
            if oldRoot.isLeftChild():
                oldRoot.parent.left = newRoot
            else:
                oldRoot.parent.right = newRoot

        newRoot.left = oldRoot
        oldRoot.parent = newRoot
        oldRoot.balanceFactor = oldRoot.balanceFactor + 1 - min(newRoot.balanceFactor, 0)
        newRoot.balanceFactor = newRoot.balanceFactor + 1 - max(oldRoot.balanceFactor, 0)

    # range query
    def lookup_r(self, start, end, node):

        if node is None:
            return None, None

        node_start = node.getStart()
        if end < node_start:
            return self.lookup_r(start, end, node.left)

        if start > node_start:
            return self.lookup_r(start, end, node.right)

        if start == node_start and end == node_start:
            return node, None

        if start == node_start and end > node_start:
            return node, self.lookup_p(end, node.right)

        if end == node_start and start < node_start:
            return self.lookup_p(start, node.left), node
        return None, None

    # point query using start point
    def lookup_p(self, start, node):
        """
        Lookup node containing data

        @param data node data object to look up
        @param parent node's parent
        @returns node and node's parent if found or None, None
        """
        if node is None:
            return None
        if start < node.getStart():
            if node.left is None:
                return None
            return self.lookup_p(start, node.left)

        elif start > node.getStart():
            if node.right is None:
                return None
            return self.lookup_p(start, node.right)
        else:
            return node


class TEfeatures:
    """index of TE annotations.
    """
    def __init__ (self):

        self.indexlist = {}
        self._length = []
        self._nameIDmap = []
        self._elements = []

    def getNames(self):
        names = []
        return self._nameIDmap

    def numInstances(self):
        return len(self._nameIDmap)

    def getElements(self):
        return self._elements

    def getStrand(self, idx):
        f_name = self._nameIDmap[idx]
        return f_name[len(f_name)-1]

    def getEleName(self, idx):
        full_name = None
        if idx >= len(self._nameIDmap) or idx < 0:
            return None
        else:
            full_name = self._nameIDmap[idx]
        if full_name is not None:
            pos = full_name.find(':')
            val = full_name[pos+1:(len(full_name)-2)]
            return val
        else:
            return None

    def getFullName(self, idx):
        if idx >= len(self._nameIDmap) or idx < 0:
            return None
        else:
            return self._nameIDmap[idx]

    def getLength(self, TE_name_idx):
        if TE_name_idx < len(self._length):
            return self._length[TE_name_idx]
        else:
            return -1

    def getFamilyID(self, chromosome, start, end):
        binID = start//TEindex_BINSIZE
        endbinID = end//TEindex_BINSIZE + 1

        if chromosome in self.indexlist:
            index = self.indexlist[chromosome]
            (node, RBnode) = index.lookup(binID, index.root, None, None)

            if node is not None and node.overlaps(binID, endbinID):
                full_name = (node.getName()).split(':')
                famid = full_name[2]
                return famid
            else:
                return None

        else:
            return None

    def findOvpTE(self, chrom, start, end):
        startbinID = start//TEindex_BINSIZE
        endbinID = end//TEindex_BINSIZE
        if start == startbinID * TEindex_BINSIZE:
            startbinID -= 1
        name_idx_list = []

        if chrom in self.indexlist:
            index = self.indexlist[chrom]
        else:
            return None

        (LBnode, RBnode) = index.lookup_r(startbinID, endbinID, index.root)

        if LBnode is not None:
            telist = LBnode.overlaps(start, end)
            name_idx_list.extend(telist)

        if RBnode is not None:
            telist = RBnode.overlaps(start, end)
            name_idx_list.extend(telist)

        return name_idx_list

    def TE_annotation(self, iv_seq):
        TEs = []
        for iv in iv_seq:
            chromosome = iv[0]
            start = iv[1]
            end = iv[2]
            strand = iv[3]
            name_idx_list = self.findOvpTE(chromosome, start, end)

            if name_idx_list is not None:
                for t in name_idx_list:
                    if strand != ".":  # stranded
                        if strand == self.getStrand(t):
                            if t not in TEs:
                                TEs.append(t)
                    else:  # not stranded
                        if t not in TEs:
                            TEs.append(t)

        return TEs

    # group by element
    def groupByEle(self, te_inst_counts):

        TEs = self.getElements()
        te_ele_counts = dict(list(zip(TEs, [0]*len(TEs))))

        for i in range(len(te_inst_counts)):
            ele_name = self.getEleName(i)

            if ele_name is None:
                sys.stderr.write("TE out of index boundary! \n")
                sys.exit(1)

            if ele_name in te_ele_counts:
                te_ele_counts[ele_name] += te_inst_counts[i]
            else:
                sys.stderr.write("TE inconsistency! "+ele_name+" \n")
                sys.exit(1)

        return te_ele_counts

    def build(self, filename):
        self.__srcfile = filename

        try:
            f = open(self.__srcfile, 'r')

        except:
            logging.error("cannot open such file %s ! \n" %(self.__srcfile))
            sys.exit(1)

        name_idx = 0
        linenum = 0
        for line in f:
            line = line.strip()

            if line.startswith("#"):
                continue

            items = line.split('\t')
            chrom = items[0]
            start = int(items[3])
            end = int(items[4])
            strand = items[6]
            items[8] = items[8].replace("; ", ";")
            desc = items[8].split(';')
            name = ""
            family_id = ""
            ele_id = ""
            class_id = ""
            tlen = end - start + 1
            linenum += 1

            for i in range(len(desc)):
                desc[i] = desc[i].replace("\"", "")
                pos = desc[i].find(" ")
                tid = desc[i][:pos]
                val = desc[i][pos+1:len(desc[i])]

                if tid == "gene_id":
                    ele_id = val
                if tid == "transcript_id":
                    name = val
                if tid == "family_id":
                    family_id = val
                if tid == "class_id":
                    class_id = val

            if ele_id == "" or name == "" or family_id == "" or class_id == "":
                sys.stderr.write(line+" \n")
                sys.stderr.write("TE GTF format error! There is no annotation at line %s. \n" % linenum)
                raise

            full_name = name+':'+ele_id+':'+family_id+':'+class_id+':'+strand
            ele_name = ele_id+':'+family_id+':'+class_id
            if ele_name not in self._elements:
                self._elements.append(ele_name)

            self._length.append(tlen)
            self._nameIDmap.append(full_name)

            if chrom in self.indexlist:
                index = self.indexlist[chrom]

                bin_startID = start//TEindex_BINSIZE
                bin_endID = end//TEindex_BINSIZE

                if start == bin_startID * TEindex_BINSIZE:
                    bin_startID -= 1
                while bin_startID <= bin_endID:
                    end_pos = min(end, (bin_startID+1) * TEindex_BINSIZE)
                    start_pos = max(start, bin_startID * TEindex_BINSIZE+1)

                    index.insert(start_pos, end_pos, name_idx)
                    bin_startID += 1

            else:
                index = BinaryTree()
                bin_startID = start//TEindex_BINSIZE
                bin_endID = end//TEindex_BINSIZE

                if start == bin_startID * TEindex_BINSIZE:
                    bin_startID -= 1

                while bin_startID <= bin_endID:
                    end_pos = min(end, (bin_startID+1) * TEindex_BINSIZE)
                    start_pos = max(start, bin_startID * TEindex_BINSIZE+1)
                    index.insert(start_pos, end_pos, name_idx)
                    bin_startID += 1

                self.indexlist[chrom] = index

            name_idx += 1

        f.close()


if __name__ == '__main__':
    try:

        main()

    except KeyboardInterrupt:
        sys.stderr.write("User interrupt! \n")
        sys.exit(0)



# licensed under the MIT License
# http://opensource.org/licenses/mit-license.php

import operator


class IntervalTree(object):
    __slots__ = ('intervals', 'left', 'right', 'center')

    def __init__(self, intervals, depth=16, minbucket=16, _extent=None, maxbucket=512):
        """\
        `intervals` a list of intervals *with start and stop* attributes.
        `depth`     the depth of the tree
        `minbucket` if any node in the tree has fewer than minbucket
                    elements, make it a leaf node
        `maxbucket` even it at specifined `depth`, if the number of intervals >
                    maxbucket, split the node, make the tree deeper.

        depth and minbucket usually do not need to be changed. if
        dealing with large numbers (> 1M) of intervals, the depth could
        be increased to 24.

        Usage:

         >>> ivals = [Interval(2, 3), Interval(1, 8), Interval(3, 6)]
         >>> tree = IntervalTree(ivals)
         >>> sorted(tree.find(1, 2))
         [Interval(2, 3), Interval(1, 8)]

        this provides an extreme and satisfying performance improvement
        over searching manually over all 3 elements in the list (like
        a sucker). 

        the IntervalTree class now also supports the iterator protocol
        so it's easy to loop over all elements in the tree:

         >>> import operator
         >>> sorted([iv for iv in tree], key=operator.attrgetter('start'))
         [Interval(1, 8), Interval(2, 3), Interval(3, 6)]


        NOTE: any object with start and stop attributes can be used
        in the incoming intervals list.
        """ 

        depth -= 1
        if (depth == 0 or len(intervals) < minbucket) and len(intervals) < maxbucket:
            self.intervals = intervals
            self.left = self.right = None
            return 

        if _extent is None:
            # sorting the first time through allows it to get
            # better performance in searching later.
            intervals.sort(key=operator.attrgetter('start'))

        left, right = _extent or (intervals[0].start, max(i.stop for i in intervals))
        center = (left + right) / 2.0

        self.intervals = []
        lefts, rights = [], []

        for interval in intervals:
            if interval.stop < center:
                lefts.append(interval)
            elif interval.start > center:
                rights.append(interval)
            else:  # overlapping.
                self.intervals.append(interval)
                
        self.left = lefts and IntervalTree(lefts,  depth, minbucket, (intervals[0].start, center)) or None
        self.right = rights and IntervalTree(rights, depth, minbucket, (center, right)) or None
        self.center = center

    def find(self, start, stop):
        """find all elements between (or overlapping) start and stop"""
        if self.intervals and not stop < self.intervals[0].start:
            overlapping = [i for i in self.intervals if i.stop >= start and i.start <= stop]
        else:
            overlapping = []

        if self.left and start <= self.center:
            overlapping += self.left.find(start, stop)

        if self.right and stop >= self.center:
            overlapping += self.right.find(start, stop)

        return overlapping

    def find_gene(self, start, stop):
        """find all elements between (or overlapping) start and stop"""
        if self.intervals and not stop < self.intervals[0].start:
            overlapping = [i.gene for i in self.intervals if i.stop >= start and i.start <= stop]
        else:
            overlapping = []

        if self.left and start <= self.center:
            overlapping += self.left.find_gene(start, stop)

        if self.right and stop >= self.center:
            overlapping += self.right.find_gene(start, stop)

        return overlapping

    def __iter__(self):
        if self.left:
            for l in self.left:
                yield l

        for i in self.intervals:
            yield i

        if self.right:
            for r in self.right:
                yield r


class Interval(object):
    def __init__(self, gene_id, start, stop):
        self.start = start
        self.stop = stop
        self.gene = gene_id

    def __repr__(self):
        return "%s (%i, %i)" % (self.gene, self.start, self.stop)



'''
Created on Oct 17, 2011

@author: yjin
'''

def base_m(m_ref1,m_ref2):
    target = []
   
    for i in range(len(m_ref1)):
        target.append((m_ref1[i]+m_ref2[i])/2)
    
    return target

def raw_count_mean(m_ref1,m_ref2,norm_ref) :
    target_ref = []
    for i in range(len(m_ref1)):
        target_ref.append(raw_sum_mean(m_ref1[i],m_ref2[i],norm_ref))
    
    return target_ref

def raw_sum_mean(m1,m2,tr_norm):
    v = 0;    
    for i in tr_norm :
        v += 1.0/i;   
    
    return v*(m1+m2)/2



def est_var(tr_fit,base_mean,norm,tr_raw_sum_mean):
    #including raw variance scv adjust
    s = 0
    target = []
    for n in norm :
        s += (1/n)**2
    
    for k in range(len(tr_fit)):
        var = tr_fit[k] * base_mean[k] * base_mean[k] * s + tr_raw_sum_mean[k]
        correct = 1.00000001 * tr_raw_sum_mean[k]
        
        if var < correct :
            target.append(correct)
        else:
            target.append(var)
    
    return target



'''
Created on Oct 13, 2011

@author: Ying Jin
'''
import sys
import os
import struct
import logging
import struct
import gzip
from array import array
from time import time
from math import ceil,floor 
from TEToolkit.Constants import BIN_SIZE,MAX_BIT, species_chrom_lengths
from TEToolkit.IO.FeatIO import FWTrackII
from TEToolkit.TEindex import *
#import rpy2.robjects as robjects


class Read :
    def __init__(self):
        self.chrom  = ""
        self.start  = -1
        self.end    = -1
        self.strand   = "" 
        self.name = ""   
        self.qual = 0
        
class GenericParser:
    """Generic Parser class.

    Inherit this to write your own parser.
    """
    def __init__ (self, filename):
        try :
            self.fhd = open(filename,'r')
        except:
            sys.exit(1)
        return

    def tsize(self):
        return

    def build_fwtrack (self):
        return 

    def __fw_parse_line (self, thisline ):
        return

    def sniff (self):
        try:
            t = self.tsize()
        except:
            self.fhd.seek(0)
            return False
        else:
            if t<=10 or t>=10000:
                self.fhd.seek(0)
                return False
            else:
                self.fhd.seek(0)
                return t
            
class SAMFile :
    '''
      short reads in sam format
      saved by chromosome
   '''
    
    
    def __init__(self, srcfile,chroms):
        
        self.__srcfile = srcfile
        self.__fileList = dict()
        self.size = 0
        self.__binTags = []
                
class BEDFile(GenericParser) :
    '''
      short reads in bed format
      saved by chromosome
    '''


    def __init__(self, srcfile):
        '''
        Constructor
        '''
        self.__srcfile = srcfile
        self.__fileList = dict()
        self.size = 0
        self.__tsize = 0
        self.__binTags = []
        self.__buildAready = False
        
        self.fhd = open(srcfile, 'r')
        #self.__seprate_by_chrom(chroms)
        #self.__seprate_by_chrom()
        
    def sameFam(self,multi_reads,teIdx,seq_name=""):
        famlist = {}
        sel_reads = []
        #sel_idx = []
        
        for r in multi_reads :
            chr = r.chrom
            start = r.start-100
            end = r.start + 100
            famID = teIdx.getFamilyID(chr,start,end)
            if famID not in famlist :
                famlist[famID] = []
            famlist[famID].append(r)
            
        num_reads = len(multi_reads)
        max_fam1 = ""
        max_fam2 = ""
        max_fam1_cnt = 0
        max_fam2_cnt = 0
        for k in famlist:
            k_len = len(famlist[k])
            if max_fam1_cnt < k_len :
                max_fam1_cnt = k_len
                max_fam1 = k
            else :
                if max_fam2_cnt < k_len :
                    max_fam2_cnt = k_len
                    max_fam2_cnt = k
        if max_fam1_cnt > 2.5 * max_fam2_cnt :
            w = 1.0/len(famlist[max_fam1])
            return famlist[max_fam1],w
        else :
            sel_reads.extend(famlist[max_fam1])
            sel_reads.extend(famlist[max_fam2])
            w = 1.0/(len(famlist[max_fam1])+len(famlist[max_fam2]))
            return sel_reads, w                
        
    def builtAready (self):
        return self.__buildAready
    
    def tsize(self):
        return self.__tsize
    
    def libsize(self):
        return self.size
    
    def build_fwtrack (self,temode):

        fwtrack = FWTrackII(filename=self.__srcfile)
        i = 0
        m = 0
        pre_seq_name = ""
        multi_reads = []
        seq_len_count = 0
        seq_len = 0
  #      cnt = 0
        strand = 0
        try:
            f = open(self.__srcfile,'r')
        except IOError:
            logging.error("open file %s error !\n" %(self.__srcfile))
            sys.exit(1)
        else:
            for line in f:                
        # Go through bed file and assign each line to corresponding file.
                
                line = line.strip()
                items = line.split('\t')
                chrname = items[0]
                
                if seq_len_count < 1000 :
                    seq_len += int(items[2]) - int(items[1])
                    seq_len_count += 1

                i+=1
                start = 0
                if i == 1000000:
                        m += 1
                        logging.info(" %d" % (m*1000000))
                        i=0
                
                if items[5] == "+" :
                        strand = 0
                        start = int(items[1])
                else :
                        strand = 1         
                        start = int(items[2])       
                w = 1.0
                if len(items) > 6 : #there is weight assigned to each alignment
                    w = float(items[6])
                #    self.size += w 
                    if temode == 'uniq' and w < 1.0 :
                       continue
                    fwtrack.add_loc(chrname,start,strand,w)
                    
                else :  
                    if pre_seq_name == "" :
                        pre_seq_name = items[3] 
                         
                    r = Read()
                    r.chrom = chrname
                    r.start = start
                    r.strand = strand
                    if pre_seq_name == items[3] :
                        multi_reads.append(r)
                    else :
                        if (temode =='uniq' and len(multi_reads) ==1) or (temode =='multi') :
                            w = 1.0/len(multi_reads)
                            for read in multi_reads :
                                fwtrack.add_loc(read.chrom,read.start,read.strand,w)
                            #self.size
                        multi_reads = []
                        pre_seq_name = items[3]
                        multi_reads.append(r)
                    
              #  else:
               #     logging.warn("Unspecified chromosome name at %s line: %s. Skip!\n" %(self.__srcfile,line))
            
            if len(multi_reads) > 0 :
              if (temode =='uniq' and len(multi_reads) ==1) or (temode =='multi') :
                
                  w = float(1.0/len(multi_reads))
                  for k in range(len(multi_reads)) :
                     read = multi_reads[k]
                
                     fwtrack.add_loc(read.chrom,read.start,read.strand,w)
                         
            f.close()               
        if seq_len_count > 0 :
            fwtrack.setTsize(int(seq_len/seq_len_count))    
        
        fwtrack.sort()
        self.__buildAready = True
       
        return fwtrack
        
    def build_fwtrack_v2 (self,teIdx):

        
        fwtrack = FWTrackII(filename=self.__srcfile)
        i = 0
        m = 0
        pre_seq_name = ""
        multi_reads = []
        seq_len_count = 0
        seq_len = 0
  #      cnt = 0
        strand = 0
        try:
            f = open(self.__srcfile,'r')
        except IOError:
            logging.error("open file %s error !\n" %(self.__srcfile))
            sys.exit(1)
        else:
            for line in f:                
        # Go through bed file and assign each line to corresponding file.
                
                line = line.strip()
                items = line.split('\t')
                chrname = items[0]
                
                if seq_len_count < 1000 :
                    seq_len += int(items[2]) - int(items[1])
                    seq_len_count += 1

                i+=1
                start = 0
                
                if i == 1000000:
                        m += 1
                        logging.info(" %d" % (m*1000000))
                        i=0
                
                if items[5] == "+" :
                        strand = 0
                        start = int(items[1])
                        end = int(items[2])
                else :
                        strand = 1         
                        start = int(items[2])
                               
                w = 1.0
                if len(items) > 6 : #there is weight assigned to each alignment
                    w = float(items[6])
                #    self.size += w 
                    fwtrack.add_loc(chrname,start,strand,w)
                    
                else :  
                    if pre_seq_name == "" :
                        pre_seq_name = items[3] 
                         
                    r = Read()
                    r.chrom = chrname
                    r.start = start
                    
                    r.strand = strand
                    if pre_seq_name == items[3] :
                        multi_reads.append(r)
                    else :
                        (sel_reads,w) = self.sameFam(multi_reads,teIdx,pre_seq_name)
                       # w = 1.0/len(multi_reads)
                        for k in range(len(sel_reads)) :
                            read = sel_reads[k]
                          
                            
                            #w = weights[k]
                            fwtrack.add_loc(read.chrom,read.start,read.strand,w)
                            #self.size
                        multi_reads = []
                        pre_seq_name = items[3]
                        multi_reads.append(r)
                    
              #  else:
               #     logging.warn("Unspecified chromosome name at %s line: %s. Skip!\n" %(self.__srcfile,line))
            
            if len(multi_reads) > 0 :
                
             # w = float(1.0/len(multi_reads))
              #for k in range(len(multi_reads)) :
               # read = multi_reads[k]
              (sel_reads,w) = self.sameFam(multi_reads,teIdx)
                       # w = 1.0/len(multi_reads)
              for k in range(len(sel_reads)) :
                            read = sel_reads[k]
                            #w = weights[k]  
                            fwtrack.add_loc(read.chrom,read.start,read.strand,w)
                         
            f.close()               
        if seq_len_count > 0 :
            fwtrack.setTsize(int(seq_len/seq_len_count))    
        
        fwtrack.sort()
        self.__buildAready = True
       
        return fwtrack                

        
   # def __seprate_by_chrom (self,chroms):
    def __seprate_by_chrom (self):
        
        timestamp = time()
        fhead = "." + str(timestamp)
      
        #chroms = chrlen_tbl.keys()
        fhandels = {}
        seq_len = 0
        seq_len_count =1
                   
        try:
            f = open(self.__srcfile,'r')
        except IOError:
            logging.error("open file %s error !\n" %(self.__srcfile))
            sys.exit(1)
        else:
            for line in f:                
        # Go through bed file and assign each line to corresponding file.
                line = line.strip()
                items = line.split('\t')
                chrname = items[0]
                if seq_len_count < 1000 :
                    seq_len += int(items[2]) - int(items[1])
                    seq_len_count += 1
                if chrname not in fhandels :
                    chrfile = fhead + chrname + ".bed"
                    fh = open(chrfile,'a')
                    self.__fileList[chrname] = chrfile
                    fhandels[chrname] = fh
                    
                    fh = fhandels[items[0]] 
                    
                    w = 1.0
                    if len(items) > 6 : #there is weight assigned to each alignment
                        w = float(items[6])
                    self.size += w 
                    fh.write(items[0]+"\t"+items[1]+"\t"+items[2]+"\t"+items[3]+"\t"+items[4]+"\t"+items[5]+"\t"+str(w)+"\n")
                    
                else :    
                
                    fh = fhandels[items[0]] 
                    #fh.write(line+"\n")
                    w = 1.0
                    if len(items) > 6 : #there is weight assigned to each alignment
                        w = float(items[6])
                    self.size += w  
                    fh.write(items[0]+"\t"+items[1]+"\t"+items[2]+"\t"+items[3]+"\t"+items[4]+"\t"+items[5]+"\t"+str(w)+"\n")
                      
              #  else:
               #     logging.warn("Unspecified chromosome name at %s line: %s. Skip!\n" %(self.__srcfile,line))
            
            f.close()   
        self.__tsize = int(seq_len/seq_len_count)      
        # Close all files.
        for fh in list(fhandels.values()):
            fh.close()
           
    def del_chrom_bed(self,chrom):
        f = self.__fileList[chrom]
        try:
            os.remove(f)
            
        except IOError :
            logging.error("cannot remove file %s !\n" %(f))
            sys.exit(1)
        
        #remove f
        
    def get_bin_rc (self,chrom,bin_size,fragsize,chrsize):
        
        reads = array('f',(0,) * int(ceil(1.0 *chrsize/bin_size)))
        sizeOfArray = len(reads)
        
        if chrom not in self.__fileList:
            logging.warn("No reads at chromosome %s. Skip!\n" %(chrom))
            return None
        else:
            fn = self.__fileList[chrom]
            try:
                fh = open(fn,'r')
            except IOError:
                logging.error("cannot open file of chromosome %s.\n" %(chrom))
                return None
            else:
                for line in fh:
                    line = line.strip()
                    items = line.split('\t')
                    if len(items) < 6 :
                        logging.warn("Format error at %s line: %s. Skip!\n" %(fn,line))
                    else:
                        start = int(items[1])
                        end = int(items[2]) 
                        strand = items[5]
                        w = 1
                        if len(items) > 6 : # weight for each tag
                            w = float(items[6])
                        # Shift read location by fragment size.
                        loc = 0
                        if strand == '+' :
                            loc = start + int(fragsize/2) -1
                        else :
                            loc = end - int(fragsize/2) -1
                        # Increment read count in bin.
                        pos = int(floor(1.0*loc/bin_size))   
                       
                        if(pos >= sizeOfArray) :
                            logging.error("%s  %d array index out of range! species may not match." %(items[0],pos))
                            sys.exit(0)
                        reads[pos] += w
                        
                fh.close()
                return reads
     

class BAMFile(GenericParser) :
    '''
      short reads in BAM format
      saved by chromosome
      The bitwise flag is made like this:
    dec    meaning
    ---    -------
    1    paired read
    2    proper pair
    4    query unmapped
    8    mate unmapped
    16    strand of the query (1 -> reverse)
    32    strand of the mate
    64    first read in pair
    128    second read in pair
    256    alignment is not primary
    512    does not pass quality check
    1024    PCR or optical duplicate
    '''
    def __init__(self, srcfile):
        '''
        Constructor
        '''
        self.__srcfile = srcfile
        self.__fileList = dict()
#        self.size = 0
#        self.__tsize = 0

        self.__buildAready = False
        
        self.fhd = gzip.open(srcfile, 'r')
        try:
            self.fhd.read(10)
        except IOError:
        # not a gzipped file
            logging.error("it's not a BAM file %s !\n" %(srcfile))
            self.fhd.close()
            sys.exit(1)
        else:
            self.fhd.seek(0)
        
        #self.__seprate_by_chrom(chroms)
        #self.__seprate_by_chrom()
        
    def sameFam(self,multi_reads,teIdx):
        famlist = {}
        sel_reads = []
        #sel_idx = []
        
        for r in multi_reads :
            chr = r.chrom
            start = r.start - 100
            end = r.end + 100
            famID = teIdx.getFamilyID(chr,start,end)
            if famID not in famlist :
                famlist[famID] = []
            famlist[famID].append(r)
            
        num_reads = len(multi_reads)
        max_fam1 = ""
        max_fam2 = ""
        max_fam1_cnt = 0
        max_fam2_cnt = 0
        for k in famlist:
            k_len = len(famlist[k])
            if max_fam1_cnt < k_len :
                max_fam1_cnt = k_len
                max_fam1 = k
            else :
                if max_fam2_cnt < k_len :
                    max_fam2_cnt = k_len
                    max_fam2_cnt = k
        
        if max_fam1_cnt > 2.5 * max_fam2_cnt :
            w = 1.0/len(famlist[max_fam1])
            return famlist[max_fam1],w
        else :
            sel_reads.extend(famlist[max_fam1])
            sel_reads.extend(famlist[max_fam2])
            w = 1.0/(len(famlist[max_fam1])+len(famlist[max_fam2]))
            return sel_reads, w                
        
    def builtAready (self):
        return self.__buildAready

    def build_fwtrack (self,temode):
        """Build FWTrackII from all lines, return a FWTrackII object.

        Note only the unique match for a tag is kept.
        """
        fwtrack = FWTrackII(filename=self.__srcfile)
        fseek = self.fhd.seek
        fread = self.fhd.read
        ftell = self.fhd.tell
        references = []
        # move to pos 4, get the length of header
        fseek(4)
        header_len =  struct.unpack('<i', fread(4))[0]
        fseek(header_len + ftell())
        # get the number of chromosome
        nc = struct.unpack('<i', fread(4))[0]
     
        for x in range(nc):
            # read each chromosome name
            nlength = struct.unpack('<i', fread(4))[0]
            chrname = fread(nlength)[:-1]
            references.append(chrname)
            # jump over chromosome size, we don't need it
            fseek(ftell() + 4)
        
        i = 0
        m = 0
        multi_reads = []
        prev_seq = ""
        seq_len = 0
        seq_len_count = 0
        while 1:
            try:
                entrylength = struct.unpack('<i', fread(4))[0]
            except struct.error:
               
                break
            (seq_name,chrid,fpos,flen,strand,qual) = self.__binary_parse(fread(entrylength))  
            if seq_len_count < 1000 :
                seq_len += flen
                seq_len_count += 1
            
                                  
            i+=1
            if i == 1000000:
                m += 1
                logging.info(" %d" % (m*1000000))
                i=0

            if fpos >= 0:
                if seq_name == prev_seq : #multi reads
                    r = Read()
                    r.chrom = references[chrid]
                    if strand == 1 :
                        r.strand = 1
                        r.start = fpos - flen
                        r.end = fpos
                    else :
                        r.start = fpos
                        r.end = fpos + flen
                        r.strand = 0
                    r.name = seq_name
                    r.weight = 0
                    r.qual = qual
                    
                    multi_reads.append(r)
                else :
                    if prev_seq != "" :
                        if (temode == 'uniq' and len(multi_reads) == 1)  or temode == 'multi':
                            w = round(1.0/len(multi_reads),2)
                            for k in range(len(multi_reads)) :
                                rr = multi_reads[k]
                                fwtrack.add_loc(rr.chrom,rr.start,rr.strand,w)
                            
                    multi_reads = []
                    prev_seq = seq_name
                    r = Read()
                    r.chrom = references[chrid]
                    if strand == 1 :
                        r.strand = 1
                        r.start = fpos - flen
                        r.end = fpos
                    else :
                        r.start = fpos
                        r.end = fpos + flen
                        r.strand = 0
                    r.name = seq_name
                    r.qual = qual
                    
                    multi_reads.append(r)
        
        
        if len(multi_reads) > 0 :
            if (temode == 'uniq' and len(multi_reads) == 1)  or temode == 'multi':
                w = round(1.0/len(multi_reads),2)
                for k in range(len(multi_reads)) :
                    rr = multi_reads[k]
                    fwtrack.add_loc(rr.chrom,rr.start,rr.strand,w)
                 
                                
        self.fhd.close()
        if seq_len_count > 0 :
            fwtrack.setTsize(int(seq_len/seq_len_count))

        self.__buildAready = True
        
        return fwtrack
    
       
    def build_fwtrack_v2 (self,teIdx):
        """Build FWTrackII from all lines, return a FWTrackII object.

        Note only the unique match for a tag is kept.
        """
        fwtrack = FWTrackII(filename=self.__srcfile)
        fseek = self.fhd.seek
        fread = self.fhd.read
        ftell = self.fhd.tell
        references = []
        # move to pos 4, get the length of header
        fseek(4)
        header_len =  struct.unpack('<i', fread(4))[0]
        fseek(header_len + ftell())
        # get the number of chromosome
        nc = struct.unpack('<i', fread(4))[0]
     
        for x in range(nc):
            # read each chromosome name
            nlength = struct.unpack('<i', fread(4))[0]
            chrname = fread(nlength)[:-1]
            references.append(chrname)
            # jump over chromosome size, we don't need it
            fseek(ftell() + 4)
        
        i = 0
        m = 0
        multi_reads = []
        prev_seq = ""
        seq_len = 0
        seq_len_count = 0
        while 1:
            try:
                entrylength = struct.unpack('<i', fread(4))[0]
            except struct.error:
               
                break
            (seq_name,chrid,fpos,flen,strand,qual) = self.__binary_parse(fread(entrylength))  
            if seq_len_count < 1000 :
                seq_len += flen
                seq_len_count += 1
            
                                  
            i+=1
            if i == 1000000:
                m += 1
                logging.info(" %d" % (m*1000000))
                i=0

            if fpos >= 0:
                if seq_name == prev_seq : #multi reads
                    r = Read()
                    r.chrom = references[chrid]
                    if strand == 1 :
                        r.strand = 1
                        r.start = fpos - flen
                        r.end = fpos
                    else :
                        r.start = fpos
                        r.end = fpos + flen
                        r.strand = 0
                    r.name = seq_name
                    r.weight = 0
                    r.qual = qual
                    
                    multi_reads.append(r)
                else :
                    if prev_seq != "" :
                        (sel_reads,w) = self.sameFam(multi_reads,teIdx)
                       # w = 1.0/len(multi_reads)
                        for k in range(len(sel_reads)) :
                            rr = sel_reads[k]
                            #w = weights[k]                        
                       # w = round(1.0/len(multi_reads),2)
                        #for k in range(len(multi_reads)) :
                         #   rr = multi_reads[k]
                            fwtrack.add_loc(rr.chrom,rr.start,rr.strand,w)
                            
                    multi_reads = []
                    prev_seq = seq_name
                    r = Read()
                    r.chrom = references[chrid]
                    if strand == 1 :
                        r.strand = 1
                        r.start = fpos - flen
                        r.end = fpos
                    else :
                        r.start = fpos
                        r.end = fpos + flen
                        r.strand = 0
                    r.name = seq_name
                    r.qual = qual
                    
                    multi_reads.append(r)
        
        
        if len(multi_reads) > 0 :
            (sel_reads,w) = self.sameFam(multi_reads,teIdx)
                       # w = 1.0/len(multi_reads)
            for k in range(len(sel_reads)) :
                 rr = sel_reads[k]
                 #w = weights[k]
            #w = round(1.0/len(multi_reads),2)
            #for k in range(len(multi_reads)) :
             #   rr = multi_reads[k]
                 fwtrack.add_loc(rr.chrom,rr.start,rr.strand,w)
                 
                                
        self.fhd.close()
        if seq_len_count > 0 :
            fwtrack.setTsize(int(seq_len/seq_len_count))

        self.__buildAready = True
        
        return fwtrack
    
    
    def sniff(self):
        if self.fhd.read(3) == "BAM":
            return True
        else:
            return False
        
    #def __seprate_by_chrom (self,chroms):
    def __seprate_by_chrom (self):
        
        timestamp = time()
        fhead = "." + str(timestamp)
      
        #chroms = chrlen_tbl.keys()
        fhandels = {}
        
        # Create chromosome files for writing.
    #    for c in chroms:
    #        chrfile = fhead + c + ".bed"
    #        fh = open(chrfile,'a')           
    #        self.__fileList[c] = chrfile
    #        fhandels[c] = fh
        
        fseek = self.fhd.seek
        fread = self.fhd.read
        ftell = self.fhd.tell
        references = []
        # move to pos 4, get the length of header
        fseek(4)
        header_len =  struct.unpack('<i', fread(4))[0]
        fseek(header_len + ftell())
        # get the number of chromosome
        nc = struct.unpack('<i', fread(4))[0]
     
        for x in range(nc):
            # read each chromosome name
            nlength = struct.unpack('<i', fread(4))[0]
            chrname = fread(nlength)[:-1]
            references.append(chrname)
            if chrname not in fhandels :
                chrfile = fhead + chrname + ".bed"
                fh = open(chrfile,'a')
                self.__fileList[chrname] = chrfile
                fhandels[chrname] = fh
            # jump over chromosome size, we don't need it
            fseek(ftell() + 4)
        
        i = 0
        m = 0
        multi_reads = []
        prev_seq = ""
        seq_len = 0
        seq_len_count = 0
        while 1:
            try:
                entrylength = struct.unpack('<i', fread(4))[0]
            except struct.error:
               
                break
            (seq_name,chrid,fpos,flen,strand,qual) = self.__binary_parse(fread(entrylength))  
            if seq_len_count < 1000 :
                seq_len += flen
                seq_len_count += 1
            
                                  
            i+=1
            if i == 1000000:
                m += 1
                logging.info(" %d" % (m*1000000))
                i=0
               # print(fpos)
               # print(seq_name)
               # print(chrid)
               # print(flen)
            if fpos >= 0:
                if seq_name == prev_seq : #multi reads
                    r = Read()
                    r.chrom = references[chrid]
                    if strand == 1 :
                        r.strand = '-'
                        r.start = fpos - flen
                        r.end = fpos
                    else :
                        r.start = fpos
                        r.end = fpos + flen
                        r.strand = '+'
                    r.name = seq_name
                    r.weight = 0
                    r.qual = qual
                    
                    multi_reads.append(r)
                else :
                    if prev_seq != "" :
                        w = 1/len(multi_reads)
                        for k in range(len(multi_reads)) :
                            rr = multi_reads[k]
                            if rr.chrom in fhandels:
                                fh = fhandels[rr.chrom] 
                                fh.write(rr.chrom+"\t"+str(rr.start)+"\t"+str(rr.end)+"\t"+rr.name+"\t"+str(rr.qual)+"\t"+rr.strand+"\t"+str(w)+"\n")
                                self.size += w    
                            else :
                                
                                logging.warn("Unspecified chromosome name at %s line: %s. Skip!\n" %(rr.chrom))
                    multi_reads = []
                    prev_seq = seq_name
                    r = Read()
                    r.chrom = references[chrid]
                    if strand == 1 :
                        r.strand = '-'
                        r.start = fpos - flen
                        r.end = fpos
                    else :
                        r.start = fpos
                        r.end = fpos + flen
                        r.strand = '+'
                    r.name = seq_name
                    r.qual = qual
                    
                    multi_reads.append(r)
        
        
        if len(multi_reads) > 0 :
            w = 1/len(multi_reads)
            for k in range(len(multi_reads)) :
                rr = multi_reads[k]
                if rr.chrom in fhandels:
                    fh = fhandels[rr.chrom] 
                    fh.write(rr.chrom+"\t"+str(rr.start)+"\t"+str(rr.end)+"\t"+rr.name+"\t"+str(rr.qual)+"\t"+rr.strand+"\t"+str(w)+"\n")
                    self.size += w 
                                
        self.fhd.close()
        self.__tsize = int(seq_len/seq_len_count)
        
        # Close all files.
        for fh in list(fhandels.values()):
            fh.close()
    
    def __binary_parse (self, data ):
        # we skip lot of the available information in data (i.e. tag name, quality etc etc)
        if not data: return (None,None,-1,-1,None,None)

        thisref = struct.unpack('<i', data[0:4])[0]
        thisstart = struct.unpack('<i', data[4:8])[0]
        (cigar, bwflag) = struct.unpack('<hh', data[12:16])
        (bin,mq_nl) = struct.unpack('<hh',data[8:12])
        name_len = mq_nl & 255
        
        seq_len = struct.unpack('<i',data[16:20])[0]
        seq_name = data[32:(32 + name_len)]
        seq_name = seq_name[:-1]
        qual = 0
        pairedEnd = False
        
        if bwflag & 4 or bwflag & 512 or bwflag & 1024:
            return (None,None, -1, -1,None,None)       #unmapped sequence or bad sequence
        if bwflag & 1:
            # paired read. We should only keep sequence if the mate is mapped
            # and if this is the left mate, all is within  the flag! 
            if not bwflag & 2:
                return (None,None, -1, -1,None,None)   # not a proper pair
            if bwflag & 8:
                return (None,None, -1, -1,None,None)   # the mate is unmapped
            p1pos = thisstart
            p2pos = struct.unpack('<i', data[24:28])[0]
            if p1pos > p2pos:
                # this pair is the farthest one, skip it
                return (None,None, -1, -1,None,None)
        # In case of paired-end we have now skipped all possible "bad" pairs
        # in case of proper pair we have skipped the rightmost one... if the leftmost pair comes
        # we can treat it as a single read, so just check the strand and calculate its
        # start position... hope I'm right!
        if bwflag & 16:
            thisstrand = 1
            thisstart = thisstart + struct.unpack('<i', data[16:20])[0]    #reverse strand should be shifted len(query) bp 
        else:
            thisstrand = 0

        return (seq_name,thisref, thisstart, seq_len,thisstrand,qual)      
        
 



from setuptools import setup

def readme():
      with open('README.rst') as f:
               return f.read()

setup(name='TEtranscripts',
      version='2.2.1',
      description='Tools for estimating differential enrichment of Transposable Elements and other highly repetitive regions',
      long_description=readme(),
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Science/Research',
          'Environment :: Console',
          'Natural Language :: English',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Topic :: Scientific/Engineering :: Bio-Informatics',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: Unix'
      ],
      keywords='TE transposable element differential enrichment',
      url='http://hammelllab.labsites.cshl.edu/software#TEtranscripts',
      author='Ying Jin, Eric Paniagua, Talitha Forcier, Oliver Tam, Molly Hammell',
      author_email='yjin@cshl.edu',
      license='GPLv3',
      packages=[
          'TEToolkit',
          'TEToolkit.IO',
          'TEToolkit.ShortRead'
      ],
      platforms=[
          'Linux',
          'MacOS'
      ],
      install_requires=[
          'argparse',
          'pysam>=0.9'
      ],
      include_package_data=True,
      zip_safe=False,
      scripts=[
          'bin/TEtranscripts',
          'bin/TEcount'
      ]
)



'''
Created on Oct 12, 2011

this code is for reading input parameters.

@author: Ying Jin
@status: 
@contact: yjin@cshl.edu
@version: 
'''
# python modules
import sys
import os
import re
import logging
import time
import gzip
import collections
from  math import log

from TEToolkit.Constants import *
from TEToolkit.ShortRead.ParseBEDFile import BEDFile,BAMFile,SAMFile


#Taken from HTSeq
class FileOrSequence( object ):
   """ The construcutor takes one argument, which may either be a string,
   which is interpreted as a file name (possibly with path), or a
   connection, by which we mean a text file opened for reading, or 
   any other object that can provide an iterator over strings 
   (lines of the file).
   
   The advantage of passing a file name instead of an already opened file
   is that if an iterator is requested several times, the file will be
   re-opened each time. If the file is already open, its lines can be read
   only once, and then, the iterator stays exhausted.      
   
   Furthermore, if a file name is passed that end in ".gz" or ".gzip"
   (case insensitive), it is transparently gunzipped.
   """
   
   def __init__( self, filename_or_sequence ):      
      self.fos = filename_or_sequence
      self.line_no = None
      
   def __iter__( self ):
      self.line_no = 1
      if isinstance( self.fos, str ):
         if self.fos.lower().endswith( ( ".gz" , ".gzip" ) ):
            lines = gzip.open( self.fos )
         else:
            lines = open( self.fos )
      else:
         lines = self.fos
      for line in lines:
         yield line
         self.line_no += 1
      if isinstance( self.fos, str ):
         lines.close()
      self.line_no = None
         
   def __repr__( self ):
      if isinstance( self.fos, str ):
         return "<%s object, connected to file name '%s'>" % (
            self.__class__.__name__, self.fos )
      else:
         return "<%s object, connected to %s >" % (
            self.__class__.__name__, repr( self.fos ) )   
            
   def get_line_number_string( self ):
      if self.line_no is None:
         if isinstance( self.fos, str ):
            return "file %s closed" % self.fos
         else:
            return "file closed"
      if isinstance( self.fos, str ):
         return "line %d of file %s" % ( self.line_no, self.fos )
      else:
         return "line %d" % self.line_no



#Taken from HTSeq
class SAM_Reader( FileOrSequence ):
   """Parse a SAM File"""

   def __iter__( self ):
      for line in FileOrSequence.__iter__( self ):
         if line.startswith( "@" ):
            continue
         try:
            (qname, flag, rname, pos, mapq, cigar, rnext, pnext, tlen, seq, trash) = line.split("\t", 10)
         except ValueError as e:
            e.args = e.args + ( self.get_line_number_string(), )
            raise
         yield (qname, int(flag), rname, int(pos)-1, mapq, cigar, rnext, pnext, seq.upper())

def read_opts(parser):
    ''' object parser contains parsed options '''
    
    args = parser.parse_args()
    
    #treatment files
    for i in range(len(args.tfiles)) :
        if not os.path.isfile(args.tfiles[i]) :
            logging.error("No such file: %s !\n" % (args.tfiles[i]))
            sys.exit(1)
        
    if not os.path.isfile(args.tinputs[0]) :
            logging.error("No such file: %s !\n" % (args.tinputs))
            sys.exit(1)
    
    #control files 
    if args.cfiles != None :      
        for i in range(len(args.cfiles)) :
            if not os.path.isfile(args.cfiles[i]) :
                logging.error("No such file: %s !\n" % (args.cfiles[i]))
                sys.exit(1)
            else :
                if args.cinputs == None :
                    logging.error("No input for control samples!\n")
                    sys.exit(1)
    else :
        args.cinputs = None
        
    if args.TEmode != 'multi' and args.TEmode != 'uniq' :
        logging.error("Does not support TE mode : %s !\n" % (args.TEmode))
    # file parser
    if args.fileformat == "BAM" :
        args.parser = BAMFile
    elif args.fileformat == "BED" :
        args.parser = BEDFile
    else :
        logging.error("Does not support such file format: %s !\n" %(args.fileformat))
        sys.exit(1)
    #window size
    if args.wsize < 0 :
        logging.error("window size should be greater than 0, default value %d was used\n" % (WIN_SIZE))
        args.wsize = WIN_SIZE 
    
    #step size
    if args.step < 0 :
        logging.error("step size should be greater than 0, default value  %d was used\n" % (STEP))
        args.step = STEP
        
    if args.step > args.wsize :
        logging.error("step should be smaller than window size,default value %d was used\n" % (STEP))
        args.step = STEP
    
    #cutoff
    if args.minread < 0 :
        args.minread = 0
    if args.minread > 20 :
        args.minread = 20
    
    #species
    if args.species[0] not in ['hg','mm','dm','tm'] :
        logging.error("species not found %s \n" %(args.species[0]))
        parser.print_help()
        sys.exit(1)
        
    args.gsize = efgsize[args.species[0]]
    args.gsize = float(args.gsize)     
    if args.species[0] == 'hg' :
        args.chrom = HS_CHROM
        args.species[0] = 'hg19'
       
        
    elif args.species[0] == 'mm' :
        args.chrom = MM_CHROM
        args.species[0] = 'mm9'
    elif args.species[0] == 'dm' :
        args.chrom = DM_CHROM
        args.species[0] = 'dm3'
    elif args.species[0] == 'tm' :
        args.chrom = TM_CHROM
    
    #normalization
    if args.norm not in ['sd','bc'] :
        logging.error("normalization method %s not supported\n" % (args.norm))
        parser.print_help()
        sys.exit(1)
    
    #p-value
    if args.pval < 0 or args.pval > 1 :
        logging.error("p-value should be a value in [0,1]\n")
        sys.exit(1)            
    args.log_pvalue = log(args.pval,10)*-10
    #gap size
    if args.gap < 0 :
        logging.error("gap size should be greater than 0, default value was used\n")
        args.gap = GAP
    
    #fragment size
    if args.fragsize < 0 :
        logging.error("fragment size should be greater than 0, default value %d was used\n" % (FRAG_SIZE))
        args.fragsize = FRAG_SIZE
    
    #output filenames
    args.dfbs = args.prj_name+"_dfbs"
    
    # logging object
    logging.basicConfig(level=(4-args.verbose)*10,
                        format='%(levelname)-5s @ %(asctime)s: %(message)s ',
                        datefmt='%a, %d %b %Y %H:%M:%S',
                        stream=sys.stderr,
                        filemode="w"
                        )
    
    args.error   = logging.critical        # function alias
    args.warn    = logging.warning
    args.debug   = logging.debug
    args.info    = logging.info
    
    cinput = None
    if args.cinputs != None:
        cinput = args.cinputs[0]
        
    args.argtxt = "\n".join((
        "# ARGUMENTS LIST:",\
        "# name = %s" % (args.prj_name),\
        "# treatment files = %s" % (args.tfiles),\
        "# control files = %s" % (args.cfiles),\
        '# treatment input = %s' % (args.tinputs[0]),\
        '# control input = %s' % (cinput),\
 #       "# window size = %d" % (args.wsize),\
        "# step = %d" % (args.step),\
 #       "# gap = %d" % (args.gap),\
        "# fragment size = %d" % (args.fragsize),\
        "# species = %s (hg:human, rn:rat, mm:mouse)" % (args.species[0]),\
        "# min read cutoff = %d" % (args.minread),\
        "# statistical model = Poisson distribution" ,\
        "# normalization = %s (sd: sequence depth, bc: bin correlation)" % (args.norm),\
        "# pvalue cutoff = %.2e" % (args.pval),\
        "# TEmode = %s " % (args.TEmode)
     #   "# TE annotation file = %s \n" % (args.TEannotation)
        ))
    
    return args  

def read_opts2(parser):
    args = parser.parse_args()
    if not os.path.isfile(args.tefile) :
        logging.error("No such file: %s !\n" %(args.tefile))
        sys.exit(1)
    if not os.path.isfile(args.gtffile) :
        logging.error("No such file: %s !\n" % (args.gtffile))
        sys.exit(1)
    # Obtain & store list of files for group 1 (e.g. treatment/mutant)
    for i in range(len(args.tfiles)) :
        if not os.path.isfile(args.tfiles[i]) :
            logging.error("No such file: %s !\n" % (args.tfiles[i]))
            sys.exit(1)
    # Obtain & store list of files for group2 (e.g. control/wildtype)
    for i in range(len(args.cfiles)) :
        if not os.path.isfile(args.cfiles[i]) :
            logging.error("No such file: %s !\n" % (args.cfiles[i]))
            sys.exit(1)
    # Identify file format for subsequent processing (parsing)
    if args.fileformat == "BAM" :
        args.parser = "BAM"
    elif args.fileformat == "SAM" :
        args.parser = "SAM"
    else :
        logging.error("Does not support such file format: %s !\n" % (args.fileformat))
        sys.exit(1)
    # What type of RNA-Seq experiment (stranded or not)
    if args.stranded not in ['forward', 'no', 'reverse'] :
        logging.error("Does not support such stranded value: %s !\n" % (args.stranded))
        sys.exit(1)

    # Method of assigning reads to annotation (gene or TE)
    if args.te_mode not in ['uniq', 'multi'] :
        logging.error("multi-mapper counting mode %s not supported\n" % (args.te_mode))
        parser.print_help()
        sys.exit(1)
        # Method of normalization (rpm or quantile)

    if args.DESeq:
        args.DEseq=True
    else:
        args.DEseq=False

    if args.norm not in ['quant','TC','DESeq_default'] :
        logging.error("normalization method %s not supported\n" % (args.norm))
        parser.print_help()
        sys.exit(1)
        # Cutoff for adjusted p-value
    if args.pval < 0 or args.pval > 1 :
        logging.error("p-value should be a value in [0,1]\n")
        sys.exit(1)
    # Cutoff for fold change
    if args.fc == 0:
        logging.error("absolute fold change ratio cannot be zero\n")
        sys.exit(1)
    elif args.fc < 0:
        args.fc = -1.0 * args.fc
    elif args.fc < 1 :
        args.fc = 1.0/args.fc
    else:
        args.fc = 1.0 * args.fc
    
    if args.sortByPos:
        args.sortByPos=True
    else:
        args.sortByPos=False

    if args.min_read < 0 :
        args.min_read = 1
    if args.numItr < 0 :
        args.numItr = 0
    if args.fragLength < 0 :
        logging.error("the fragment length cannot be negative. \n")
        sys.exit(1)
    if args.minL < 0 :
        logging.error("the minimum fragment length cannot be negative. \n")
        sys.exit(1)
    if args.maxL < 0 :
        logging.error("the maximum fragment length cannot be negative. \n")
        sys.exit(1)

    # Level of logging for tool
    logging.basicConfig(level=(4 - args.verbose) * 10,
        format='%(levelname)-5s @ %(asctime)s: %(message)s ',
        datefmt='%a, %d %b %Y %H:%M:%S',
    stream=sys.stderr, filemode="w")
    
    args.error = logging.critical        # function alias
    args.warn = logging.warning
    args.debug = logging.debug
    args.info = logging.info
    
    args.argtxt = "\n".join(("# ARGUMENTS LIST:", \
                "# name = %s" % (args.prj_name), \
                "# treatment files = %s" % (args.tfiles), \
                "# control files = %s" % (args.cfiles), \
                "# GTF file = %s " % (args.gtffile), \
                "# TE file = %s " % (args.tefile), \
                "# multi-mapper mode = %s " % (args.te_mode), \
                "# stranded = %s\n" % (args.stranded)
    ))
    if args.DESeq:
        args.argtxt = args.argtxt + "\n".join(("# differential analysis using DESeq", \
                    "# normalization = %s (rpm: Reads Per Million mapped; quant: Quantile normalization)" % (args.norm), \
                    "# FDR cutoff = %.2e" % (args.pval), \
                    "# fold-change cutoff = %5.2f" % (args.fc), \
                    "# read count cutoff = %d" % (args.min_read), \
                    "# number of iteration = %d" % (args.numItr), \
                    "# Alignments grouped by read ID = %s\n" % (not args.sortByPos)
        ))
    else:
        args.argtxt = args.argtxt + "\n".join(("# differential analysis using DESeq2", \
                    "# normalization = DESeq2_default", \
                    "# FDR cutoff = %.2e" % (args.pval), \
                    "# fold-change cutoff = %5.2f" % (args.fc), \
                    "# read count cutoff = %d" % (args.min_read), \
                    "# number of iteration = %d" % (args.numItr), \
                    "# Alignments grouped by read ID = %s\n" % (not args.sortByPos)
        ))
    return args 

def read_opts3(parser):
    args = parser.parse_args()
    if not os.path.isfile(args.tefile) :
        logging.error("No such file: %s !\n" %(args.tefile))
        sys.exit(1)
    if not os.path.isfile(args.gtffile) :
        logging.error("No such file: %s !\n" % (args.gtffile))
        sys.exit(1)
    # Check for BAM file
    if not os.path.isfile(args.bam) :
        logging.error("No such file: %s !\n" % (args.bam))
        sys.exit(1)
    # Identify file format for subsequent processing (parsing)
    if args.fileformat == "BAM" :
        args.parser = "BAM"
    elif args.fileformat == "SAM" :
        args.parser = "SAM"
    else :
        logging.error("Does not support such file format: %s !\n" % (args.fileformat))
        sys.exit(1)
    # What type of RNA-Seq experiment (stranded or not)
    if args.stranded not in ['forward', 'no', 'reverse'] :
        logging.error("Does not support such stranded value: %s !\n" % (args.stranded))
        sys.exit(1)

    # Method of assigning reads to annotation (gene or TE)
    if args.te_mode not in ['uniq', 'multi'] :
        logging.error("multi-mapper counting mode %s not supported\n" % (args.te_mode))
        parser.print_help()
        sys.exit(1)
    
    if args.sortByPos:
        args.sortByPos=True
    else:
        args.sortByPos=False

    if args.numItr < 0 :
        args.numItr = 0
    if args.fragLength < 0 :
        logging.error("the fragment length cannot be negative. \n")
        sys.exit(1)
    if args.minL < 0 :
        logging.error("the minimum fragment length cannot be negative. \n")
        sys.exit(1)
    if args.maxL < 0 :
        logging.error("the maximum fragment length cannot be negative. \n")
        sys.exit(1)

    # Level of logging for tool
    logging.basicConfig(level=(4 - args.verbose) * 10,
        format='%(levelname)-5s @ %(asctime)s: %(message)s ',
        datefmt='%a, %d %b %Y %H:%M:%S',
        stream=sys.stderr, filemode="w")
    
    args.error = logging.critical        # function alias
    args.warn = logging.warning
    args.debug = logging.debug
    args.info = logging.info
    
    args.argtxt = "\n".join(("# ARGUMENTS LIST:", \
        "# name = %s" % (args.prefix), \
        "# BAM file = %s" % (args.bam), \
        "# GTF file = %s " % (args.gtffile), \
        "# TE file = %s " % (args.tefile), \
        "# multi-mapper mode = %s " % (args.te_mode), \
        "# stranded = %s " % (args.stranded), \
        "# number of iteration = %d" % (args.numItr), \
        "# Alignments grouped by read ID = %s\n" % (not args.sortByPos)
    ))
    return args 

def read_chrlen_tbl(chrfile,error,info):
    ''' read in chrom_size file '''
 
    if not os.path.isfile(chrfile) :
        error("No such file: %s !\n" % (chrfile))
        sys.exit(1)
    try:
        f = open(chrfile,'r')
    except IOError :
        error("open file %s error !\n" %(chrfile))
        sys.exit(1)
    else :
        chrlen_map = dict()    # hash table for chromosome length.
        cnt = 0
        for line in f :
            cnt += 1
            line = line.strip()
            items = line.split('\t')    # skip empty line.
            if len(items) < 2 :
                info("Insufficient chromosome information at % s, line: %s. Skip!\n" % (chrfile, line))
            if re.match('^(c|C)hr', items[0]) and re.match('^[0-9]+$', items[1]) :
                chrlen_map[items[0]] = int(items[1])
            else :
                info("Format error at %s, line %d: %s. Skip!\n" % (chrfile,cnt,line))
        f.close()
        
    return chrlen_map


def read_short_reads(samples,parser,TEmode):
    '''read short reads from single or multple samples and stored in short read objects '''
    
    shortReads = []
 #   chroms = chrlen_tbl.keys()
    for i in range(len(samples)) :
        s = samples[i]
        if not os.path.isfile(s) :
            logging.error("No such file %s !\n" %(s))
            sys.exit(1)
        logging.info("reading sample file %s ...\n" %(s))
        time.sleep(1)

        #sbed = __bam2bed(s,0,error)
        #b = BEDFile(sbed,chroms)
        #b = parser(s,chroms)
        b = parser(s)
        t = b.build_fwtrack(TEmode)
        shortReads.append(t)
   
    return shortReads

def read_short_reads_sameFam(samples,parser,teIdx):
    '''read short reads from single or multple samples and stored in short read objects '''
    
    shortReads = []
 #   chroms = chrlen_tbl.keys()
    for i in range(len(samples)) :
        s = samples[i]
        if not os.path.isfile(s) :
            logging.error("No such file %s !\n" %(s))
            sys.exit(1)
        logging.info("reading sample file %s ...\n" %(s))
        time.sleep(1)

        #sbed = __bam2bed(s,0,error)
        #b = BEDFile(sbed,chroms)
        #b = parser(s,chroms)
        b = parser(s)
        t = b.build_fwtrack_v2(teIdx)
        shortReads.append(t)
   
    return shortReads
   
#def read_alignments(samples,chrlen_tbl,parser):
#    '''read alignments from single or multple SAM or BAM files '''
    
#    shortReads = []
#    chroms = chrlen_tbl.keys()
#    for i in range(len(samples)) :
#        s = samples[i]
#        if not os.path.isfile(s) :
#            logging.error("No such file %s !\n" %(s))
#            sys.exit(1)
#        logging.info("reading treatment sample file %s ...\n" %(s))
#        time.sleep(1)

        #sbed = __bam2bed(s,0,error)
        #b = BEDFile(sbed,chroms)
        #b = parser(s,chroms)
#        shortReads.append(b)
   
#    return shortReads     
   
def __bam2bed(sample,pairend,error):
    res = sample + ".bed"
    if pairend == 0 : #single end
        try:
            os.system("bamToBED -ed -i sample >res")
            res = __assignWeight(sample,".bed",error)
        except :
            error("file format error %s !\n" %(sample))
            sys.exit(0)
        
    else :
        try:
            os.system("bamToBED -bedpe -i sample >res")
            res = __assignWeight(sample,".bed",error)
        except :
            error("file format error %s !\n" %(sample))
            sys.exit(0)
    
    
    return res 

def __assignWeight(sample,suffix,error):

    src = sample + suffix
    dest = sample + ".bal.bed"

    lines = []
    cur_seqid = "-1"
    multi_num = 0
    
    try:
        f = open(src,'r')
        of = open(dest,'w')
    except IOError :
        error("open file %s error !\n" %(src))
        sys.exit(1)
    else :       
        for line in f :
            line = line.strip()
            arr = line.split('\t')
            if cur_seqid == arr[3] :
                lines.append(line)
                multi_num += 1
            else :
        
                if multi_num > 0 :
                    val = 1/multi_num
                    for record in lines :
                        of.write(record + "\t" + val + "\n")
                lines.clear()
                lines.append(line)
                cur_seqid = arr[3]
                multi_num=1
    f.close()

    if multi_num > 0 :
        val = 1/multi_num
        for record in lines :
            of.write(record + "\t" + val +"\n")

    of.close()
    return dest
    

